# Standard Library
import typing as t

# Pyramid
from pyramid.registry import Registry
from pyramid.request import Request
from zope.interface import implementer

# SQLAlchemy
from sqlalchemy.orm import Session
from sqlalchemy.orm.attributes import flag_modified

import authomatic
from authomatic.core import LoginResult

# Websauna
from websauna.system.user.events import UserCreated
from websauna.system.user.interfaces import ISocialLoginMapper
from websauna.system.user.interfaces import IUserModel
from websauna.utils.time import now


class NotSatisfiedWithData(Exception):
    """Raised when social media login cannot proceed due to incomplete provided information.

    E.g. we need email to map the user, but the Facebook doesn't give us email because user doesn't grant the permission.
    """


@implementer(ISocialLoginMapper)
class EmailSocialLoginMapper:
    """Base class for mapping internal users to social network (OAuth) users."""

    def __init__(self, registry: Registry, provider_id: str):
        """Create a mapper.

        :param registry: Pyramid configuration used to drive this mapper. Subclasses might want to query this when they create users.
        :param provider_id: String id we use to refer to this authentication source in the configuration file and in the database.
        """
        #: This is the string we use to map configuration for the
        self.provider_id = provider_id
        self.registry = registry

    def activate_user(request, dbsession: Session, user: IUserModel):
        """Checks to perform when the user becomes a valid user for the first time.

        If this user has already started sign up process through email we need to cancel that.
        """
        user.activated_at = now()

        # Cancel any pending email activations if the user chooses the option to use social media login
        if user.activation:
            dbsession.delete(user.activation)

    def update_first_login_social_data(self, user: IUserModel, data: dict):
        """Set the initial data on the user model.

        When the user logs in from a social network for the first time (no prior logins with this email before) we fill in blanks in the user model with incoming data.
        Default action is not to set any items.

        :param user: User object.
        :param data: Normalized data.
        """
        pass

    def update_every_login_social_data(self, user: IUserModel, data: dict):
        """Update internal user data on every login.

        Bt default, sets user.user_data["social"]["facebook"] or user.user_data["social"]["yoursocialnetwork"] to reflect the raw data given us by ``import_social_media_user()``.

        :param user: User object.
        :param data: Normalized data.
        """
        # Non-destructive update - don't remove values which might not be present in the new data
        user.user_data["social"][self.provider_id] = user.user_data["social"].get(self.provider_id) or {}
        user.user_data["social"][self.provider_id].update(data)

        # Because we are doing direct
        flag_modified(user, "user_data")

    def create_blank_user(self, user_model: t.Callable[..., IUserModel], dbsession: Session, email: str) -> IUserModel:
        """Create a new blank user instance as we could not find matching user with the existing details.

        :param user_model: Class to be used for user creation.
        :param dbsession: SQLAlchemy Session object.
        :param email: User email.
        :return: Newly created user.
        """
        user = user_model(email=email)
        dbsession.add(user)
        dbsession.flush()
        user.username = user.generate_username()
        user.registration_source = self.provider_id
        user.activated_at = now()
        return user

    def get_existing_user(self, user_model: t.Callable[..., IUserModel], dbsession: Session, email: str) -> IUserModel:
        """Check if we have a matching user for the email already.

        :param user_model: User class to query for the user with given email.
        :param dbsession: SQLAlchemy Session object.
        :param email: User email.
        :return: User object.
        """
        user = dbsession.query(user_model).filter_by(email=email).first()
        return user

    def get_or_create_user_by_social_medial_email(self, request: Request, user: authomatic.core.User) -> IUserModel:
        """Given a user information returned by Authomatic, return an existing User object from our database or create one if it does not exists.

        :param request: Pyramid request object.
        :param user: A user returned by Authomatic.
        :return: User object.
        """
        User = self.registry.queryUtility(IUserModel)
        dbsession = request.dbsession
        imported_data = self.import_social_media_user(user)
        email = imported_data['email']
        user = self.get_existing_user(User, dbsession, email)
        if not user:
            user = self.create_blank_user(User, dbsession, email)
            request.registry.notify(UserCreated(request, user))
            self.update_first_login_social_data(user, imported_data)
            user.first_login = True
        else:
            user.first_login = False
        self.activate_user(dbsession, user)
        self.update_every_login_social_data(user, imported_data)
        return user


class FacebookMapper(EmailSocialLoginMapper):
    """Map and login Facebook OAuth users to internal users.

    You must have application created in developers.facebook.com

    The application must have its consumer_key and consumer_secret configured in the secrets config file.

    For testing: The application must have one Web site platform configured in developers.facebook.com, pointing to http://localhost:8521/ and Valid OAuth redirect URLs to http://localhost:8521/login/facebook
    """

    def import_social_media_user(self, user: authomatic.core.User) -> dict:
        """Map Authomatic user information to a dictionary.

        :param user: A user returned by Authomatic.
        :return: Mapping from authomatic.core.User.
        """
        return {
            "country": user.country,
            "timezone": user.timezone,
            "gender": user.gender,
            "first_name": user.first_name,
            "last_name": user.last_name,
            "full_name": user.name,
            "link": user.link,
            "birth_date": user.birth_date,
            "city": user.city,
            "postal_code": user.postal_code,
            "email": user.email,
            "id": user.id,
            "nickname": user.nickname,
            # "address": user.address,
        }

    def update_first_login_social_data(self, user: IUserModel, data: dict):
        """Update internal user data on every login.

        :param user: User object.
        :param data: Normalized data.
        """
        super(FacebookMapper, self).update_first_login_social_data(user, data)
        if not user.full_name and data.get("full_name"):
            user.full_name = data["full_name"]

    def capture_social_media_user(self, request: Request, result: LoginResult) -> IUserModel:
        """Extract social media information from the Authomatic login result in order to associate the user account.

        :param request: Pyramid request.
        :param result: Login result from Authomatic.
        :return: User object.
        """
        assert not result.error
        # Facebook specific Authomatic call to fetch more user data from the Facebook provider
        # https://github.com/peterhudec/authomatic/issues/112
        result.user.provider.user_info_url = 'https://graph.facebook.com/me?fields=id,email,name,first_name,last_name,gender,link,timezone,verified'
        result.user.update()

        # Make user Facebook user looks somewhat legit
        assert result.user.credentials
        assert result.user.id

        if not result.user.email:
            # We cannot login if the Facebook doesnt' give us email as we use it for the user mapping
            # This can also happen when you have not configured Facebook app properly in the developers.facebook.com
            raise NotSatisfiedWithData("Email address is needed in order to user this service and we could not get one from your social media provider. Please try to sign up with your email instead.")

        return self.get_or_create_user_by_social_medial_email(request, result.user)


class GoogleMapper(EmailSocialLoginMapper):
    """Map and login Google OAuth users to internal users.

    See :ref:`google-auth`.
    """

    def import_social_media_user(self, user: authomatic.core.User):
        """Map Authomatic user information to a dictionary.

        ref: http://peterhudec.github.io/authomatic/reference/providers.html#authomatic.providers.oauth2.Google

        :param user: A user returned by Authomatic.
        :return: Mapping from authomatic.core.User.
        """
        return {
            "email": user.email,
            "first_name": user.first_name,
            "last_name": user.last_name,
            "full_name": user.name,
            "locale": user.locale,
            "picture": user.picture,
            "email_verified": user.email_verified,
        }

    def update_first_login_social_data(self, user: IUserModel, data: dict):
        """Set the initial data on the user model.

        :param user: User object.
        :param data: Normalized data.
        """
        super(GoogleMapper, self).update_first_login_social_data(user, data)
        if not user.full_name and data.get("full_name"):
            user.full_name = data["full_name"]

    def capture_social_media_user(self, request: Request, result: LoginResult) -> IUserModel:
        """Extract social media information from the Authomatic login result in order to associate the user account.

        :param request: Pyramid request.
        :param result: Login result from Authomatic.
        :return: User object.
        """
        assert not result.error

        result.user.update()
        # Make user we got some meaningful input from the user_info_url
        assert result.user.credentials

        if not result.user.email_verified:
            raise NotSatisfiedWithData("User account email is not verified.")

        if not result.user.email:
            # We cannot login if the Facebook doesnt' give us email as we use it for the user mapping
            # This can also happen when you have not configured Facebook app properly in the developers.facebook.com
            raise NotSatisfiedWithData("Email address is needed in order to user this service and we could not get one from your social media provider. Please try to sign up with your email instead.")

        return self.get_or_create_user_by_social_medial_email(request, result.user)


class TwitterMapper(EmailSocialLoginMapper):
    """Map Twitter OAuth users to internal users.

    See :ref:`twitter-auth`.
    """

    @staticmethod
    def _x_user_parser(user: authomatic.core.User, data: dict) -> authomatic.core.User:
        """Monkey patched authomatic.providers.oauth1.Twitter._x_user_parser."""

        # dict_keys(['has_extended_profile', 'profile_use_background_image', 'time_zone', 'profile_image_url', 'default_profile_image', 'lang', 'id', 'id_str', 'profile_sidebar_border_color', 'profile_image_url_https', 'name', 'screen_name', 'utc_offset', 'email', 'location', 'friends_count', 'follow_request_sent', 'contributors_enabled', 'notifications', 'description', 'profile_link_color', 'profile_background_tile', 'followers_count', 'profile_background_image_url_https', 'is_translator', 'is_translation_enabled', 'profile_background_color', 'translator_type', 'profile_text_color', 'created_at', 'geo_enabled', 'profile_sidebar_fill_color', 'verified', 'profile_background_image_url', 'statuses_count', 'following', 'url', 'status', 'entities', 'profile_banner_url', 'default_profile', 'protected', 'favourites_count', 'listed_count'])
        user.data = data
        return user

    def import_social_media_user(self, user: authomatic.core.User):
        """Map Authomatic user information to a dictionary.

        Pass-through Twitter auth data to user_data['social']['twitter']

        :param user: A user returned by Authomatic.
        :return: Mapping from authomatic.core.User.
        """
        return user.data

    def update_first_login_social_data(self, user: IUserModel, data: dict):
        """Set the initial data on the user model.

        :param user: User object.
        :param data: Normalized data.
        """
        super(TwitterMapper, self).update_first_login_social_data(user, data)
        if not user.full_name and data.get("name"):
            user.full_name = data["name"]

    def capture_social_media_user(self, request: Request, result: LoginResult) -> IUserModel:
        """Extract social media information from the Authomatic login result in order to associate the user account.

        :param request: Pyramid request.
        :param result: Login result from Authomatic.
        :return: User object.
        """
        assert not result.error

        # We need to pass include_email=true
        # https://dev.twitter.com/rest/reference/get/account/verify_credentials
        result.provider.user_info_url = "https://api.twitter.com/1.1/account/verify_credentials.json?include_email=true"

        result.user.update()

        # Make user we got some meaningful input from the user_info_url
        assert result.user.credentials

        if not result.user.email:
            # We cannot login if the Facebook doesnt' give us email as we use it for the user mapping
            # This can also happen when you have not configured Facebook app properly in the developers.facebook.com
            raise NotSatisfiedWithData("Email address is needed in order to user this service and we could not get one from your social media provider. Please try to sign up with your email instead.")

        return self.get_or_create_user_by_social_medial_email(request, result.user)



"""Test accessing Redis."""
# Pyramid
from pyramid import testing
from pyramid.httpexceptions import HTTPOk

import pytest
from webtest import TestApp

# Websauna
from websauna.system.core.redis import create_redis
from websauna.system.core.redis import get_redis


def redis_test(request):
    redis = get_redis(request)
    redis.set("foo", b"bar")
    assert redis.get("foo") == b"bar"

    redis.set("foo", "ÅÄÖ")
    assert redis.get("foo").decode("utf-8") == "ÅÄÖ"

    return HTTPOk()


@pytest.fixture()
def app(request):
    """py.test fixture to set up a dummy app for Redis testing.

    :param request: pytest's FixtureRequest (internal class, cannot be hinted on a signature)
    """

    config = testing.setUp()
    config.add_route("home", "/")
    config.add_route("redis_test", "/redis_test")
    config.add_view(redis_test, route_name="redis_test")

    # same is in test.ini
    config.registry.settings["redis.sessions.url"] = "redis://localhost:6379/14"

    def teardown():
        testing.tearDown()

    config.registry.redis = create_redis(config.registry)

    app = TestApp(config.make_wsgi_app())
    return app


def test_access_redis(app: TestApp):
    """Simple check we can corrcetly access Redis with text encodings."""

    resp = app.get("/redis_test")
    assert resp.status_code == 200



"""Default home view for test runs."""
# Pyramid
from pyramid.view import view_config


@view_config(route_name='home', renderer='core/home.html')
def home(request):
    """The default home view of Websauna.

    You should really never see this is, as this view should be only active during Websauna test run and dev server.
    """
    return {}



"""Websauna template filters ."""
# Standard Library
import datetime
import json
import typing as t

# Pyramid
from jinja2 import Markup
from jinja2 import contextfilter
from pyramid.config import Configurator
from pyramid.renderers import render
from pyramid.threadlocal import get_current_request

import arrow
from arrow import Arrow
from pytz import timezone

# Websauna
from websauna.system.admin.utils import get_admin_url_for_sqlalchemy_object
from websauna.system.core.panel import render_panel as _render_panel
from websauna.utils import html
from websauna.utils import slug


@contextfilter
def uuid_to_slug(jinja_ctx, context, **kw):
    """Convert UUID object to a base64 encoded slug.

    Example:

    .. code-block:: html+jinja

        {% for question in latest_question_list %}
            <li>
              <a href="{{ route_url('details', question.uuid|uuid_to_slug) }}">
                {{ question.question_text }}
              </a>
            </li>
        {% endfor %}

    """
    return slug.uuid_to_slug(context)


@contextfilter
def admin_url(jinja_ctx, model_instance, *elements, **kw):
    """Link to model in admin interface.

    Takes an :term:`SQLAlchemy` model instance as a filter argument and resolves its admin page. This requires that a model admin has been correctly registered for SQLAlchemy model.

    Example:

    .. code-block:: html+jinja

        {% if request.user and request.user.is_admin %}
            <li>
              <a class="btn btn-danger" href="{{ choice|admin_url("edit") }}">
                Edit in admin
              </a>
            </li>
        {% endif %}

    Another example:

    .. code-block:: html+jinja

        <li>
          <a class="btn btn-danger" href="{{ choice|admin_url }}">
            View in admin
          </a>
        </li>
    """
    request = jinja_ctx.get('request')
    if not request:
        raise RuntimeError("Not rendered with request")

    if elements:
        view_name = elements[0]
    else:
        view_name = None

    link = get_admin_url_for_sqlalchemy_object(request.admin, model_instance, view_name=view_name)
    return link


@contextfilter
def filter_datetime(jinja_ctx, context, **kw):
    """Format datetime in a certain timezone."""
    now = context

    if not now:
        return ""

    tz = kw.get("timezone", None)
    if tz:
        tz = timezone(tz)
    else:
        tz = datetime.timezone.utc

    locale = kw.get("locale", "en_US")

    arrow = Arrow.fromdatetime(now, tzinfo=tz)

    # Convert to target timezone
    tz = kw.get("target_timezone")
    if tz:
        arrow = arrow.to(tz)
    else:
        tz = arrow.tzinfo

    format = kw.get("format", "YYYY-MM-DD HH:mm")

    text = arrow.format(format, locale=locale)

    if kw.get("show_timezone"):
        text = text + " ({})".format(tz)

    return text


@contextfilter
def arrow_format(jinja_ctx, context, *args, **kw):
    """Format datetime using Arrow formatter string.

    Context must be a time/datetime object.

    :term:`Arrow` is a Python helper library for parsing and formatting datetimes.

    Example:

    .. code-block:: html+jinja

        <li>
          Offer created at {{ offer.created_at|arrow_format('YYYYMMDDHHMMss') }}
        </li>

    `See Arrow formatting <http://crsmithdev.com/arrow/>`__.
    """
    assert len(args) == 1, "We take exactly one formatter argument, got {}".format(args)
    assert isinstance(context, (datetime.datetime, datetime.time)), "Got context {}".format(context)
    a = arrow.get(context)
    return a.format(fmt=args[0])


@contextfilter
def friendly_time(jinja_ctx, context, **kw):
    """Format timestamp in human readable format.

    * Context must be a datetimeobject

    * Takes optional keyword argument timezone which is a timezone name as a string. Assume the source datetime is in this timezone.
    """
    now = context
    if not now:
        return ""

    tz = kw.get("source_timezone", None)
    if tz:
        tz = timezone(tz)
    else:
        tz = datetime.timezone.utc

    # Make relative time between two timestamps
    now = now.astimezone(tz)
    arrow = Arrow.fromdatetime(now)
    other = Arrow.fromdatetime(datetime.datetime.utcnow())
    return arrow.humanize(other)


@contextfilter
def escape_js(jinja_ctx, context, **kw):
    """Make JSON strings to safe to be embedded inside <script> tag."""
    markup = Markup(html.escape_js(context))
    return markup


@contextfilter
def to_json(jinja_ctx, context, safe=True):
    """Converts Python dict to JSON, safe to be placed inside <script> tag.

    Example:

    .. code-block:: html+jinja

            {#
              Export server side generated graph data points
              to Rickshaw client side graph rendering
            #}
            {% if graph_data %}
              <script>
                window.graphDataJSON = "{{ graph_data|to_json }}";
              </script>
            {% endif %}

    :param context: Takes Python dictionary as input

    :param safe: Set to False to not to run ``escape_js()`` on the resulting JSON. True by default.

    :return: JSON string to be included inside HTML code
    """
    json_ = json.dumps(context)
    if safe:
        return escape_js(jinja_ctx, json_)
    else:
        return json_


@contextfilter
def render_panel(jinja_ctx, context, name, **kwargs):
    """Render a panel inline in a template.

    Allows placing :ref:`admin panels <admin-panel>` in templates directly.

    Example how to include panel at the top of admin CRUD listing template:

    .. code-block:: html+jinja

        {% block title %}

          <h1>{{title}}</h1>

          {{ context|render_panel(name="admin_panel", controls=False) }}

        {% endblock %}

    :param context: Any resource object, like ModelAdmin instance

    :param name: registered panel name, like ``admin_panel``

    :param kwargs: Passed to the panel function as is

    :return: HTML string of the rendered panel
    """
    request = jinja_ctx.get('request') or get_current_request()
    return _render_panel(context, request, name="admin_panel", **kwargs)


@contextfilter
def timestruct(jinja_ctx, context, **kw):
    """Render both humanized time and accurate time.

    * show_timezone

    * target_timezone

    * source_timezone

    * format
    """
    if not context:
        return ""

    assert type(context) in (datetime.datetime, datetime.time,)

    request = jinja_ctx.get('request') or get_current_request()
    if not jinja_ctx:
        return ""

    kw = kw.copy()
    kw["time"] = context
    kw["format"] = kw.get("format") or "YYYY-MM-DD HH:mm"

    return Markup(render("core/timestruct.html", kw, request=request))


@contextfilter
def from_timestamp(jinja_ctx, context, **kw):
    """Convert UNIX datetime to timestamp.

    Example:

    .. code-block:: html+jinja

        <p>
            Prestodoctor license expires: {{ prestodoctor.recommendation.expires|from_timestamp(timezone="US/Pacific")|friendly_time }}
        </p>

    :param context: UNIX timestamps as float as seconds since 1970
    :return: Python datetime object
    """

    tz = kw.get("timezone")
    assert tz, "You need to give an explicitimezone when converting UNIX times to datetime objects"

    # From string to object
    tz = timezone(tz)
    ct = datetime.datetime.fromtimestamp(context, tz=tz)
    return ct


def include_filter(config: Configurator, name: str, func: t.Callable, renderers=(".html", ".txt",)):
    """Register a new Jinja 2 template filter function.

    Example::

        import jinja2

        @jinja2.contextfilter
        def negative(jinja_ctx:jinja2.runtime.Context, context:object, **kw):
            '''Output the negative number.

            Usage:

                {{ 3|neg }}

            '''
            neg = -context
            return neg


    Then in your initialization:::

        include_filter(config, "neg", negative)

    :param config: Pyramid configurator

    :param name: Filter name in templates

    :param func: Python function which is the filter

    :param renderers: List of renderers where the filter is made available

    """

    def _include_filter(name, func):

        def deferred():
            for renderer_name in renderers:
                env = config.get_jinja2_environment(name=renderer_name)
                assert env, "Jinja 2 not configured - cannot add filters"
                env.filters[name] = func

        # Because Jinja 2 engine is not initialized here, only included here, we need to do template filter including asynchronously
        config.action('pyramid_web-include-filter-{}'.format(name), deferred, order=1)

    _include_filter(name, func)


def includeme(config):
    include_filter(config, "uuid_to_slug", uuid_to_slug)
    include_filter(config, "friendly_time", friendly_time)
    include_filter(config, "datetime", filter_datetime)
    include_filter(config, "escape_js", escape_js)
    include_filter(config, "timestruct", timestruct)
    include_filter(config, "to_json", to_json)
    include_filter(config, "from_timestamp", from_timestamp)
    include_filter(config, "arrow_format", arrow_format)
    include_filter(config, "render_panel", render_panel)
    include_filter(config, "admin_url", admin_url)



"""Listing views."""
# Standard Library
import typing as t

# Websauna
from websauna.system.http import Request

from . import Resource


class Column:
    """Define a column in a listing."""

    header_template = "crud/column_header.html"
    body_template = "crud/column_body.html"
    navigate_view_name = None

    #: Callback get_navigate_url(request, resource) to resolve the link where the item this column should point to
    navigate_url_getter = None
    getter = None

    #: Arrow formatting string
    format = "MM/DD/YYYY HH:mm"

    def __init__(
            self,
            id: str,
            name: t.Optional[str]=None,
            renderer=None,
            header_template: t.Optional[str]=None,
            body_template: t.Optional[str]=None,
            getter: t.Optional[t.Callable]=None,
            format: t.Optional[str]=None,
            navigate_view_name=None, navigate_url_getter=None):
        """Initialize Column.

        :param id: Column id. Must match field id on the model
        :param name: Column name, to be displayed on the header of the table.
        :param renderer:
        :param header_template: Path to header template to be used.
        :param body_template: Path to body template to be used.
        :param getter: func(view instance, object) - extract value for this column for an object
        :param format: Format to be applied to the value. i.e: "MM/DD/YYYY HH:mm" for a date value.
        :param navigate_url_getter: callback(request, resource) to generate the target URL if the contents of this cell is clicked
        :param navigate_view_name: If set, make this column clickable and navigates to the traversed name. Options are "show", "edit", "delete"
        """
        self.id = id
        self.name = name
        self.renderer = renderer
        self.getter = getter

        if format:
            self.format = format
        if header_template:
            self.header_template = header_template
        if body_template:
            self.body_template = body_template
        if navigate_view_name:
            self.navigate_view_name = navigate_view_name
        if navigate_url_getter:
            self.navigate_url_getter = navigate_url_getter

    def get_value(self, view: t.Any, obj: t.Any):
        """Extract value from the object for this column.

        Called in listing body.

        :param view: View class calling us
        :param obj: The object the list is iterating. Usually an SQLAlchemy model instance.
        """
        if self.getter:
            val = self.getter(view, self, obj)
        else:
            val = getattr(obj, self.id)

        if val is None:
            val = ''
        return val

    def get_navigate_target(self, resource: Resource, request: Request):
        """Get URL where clicking the link in the listing should go.

        By default, navigate to "show" view of the resource.

        :param resource: Traversal context
        :param request: Current HTTP Request.
        """
        return resource

    def get_navigate_url(self, resource: Resource, request: Request, view_name: t.Optional[str]=None):
        """Get the link where clicking this item should take the user.

        By default, navigate to "show" view of the resource.

        TODO: Switch resource/request argument order.
        :param resource: Traversal context
        :param request: Current HTTP Request.
        :param view_name: Override class' ``navigate_view_name``.
        """
        if self.navigate_url_getter:
            return self.navigate_url_getter(request, resource)

        if not self.navigate_view_name:
            return None

        target = self.get_navigate_target(resource, request)

        if not target:
            return None

        view_name = view_name or self.navigate_view_name

        return request.resource_url(target)


class StringPresentationColumn(Column):
    """Renders the default string presentation of the object.

    You can change the stringify method::

        StringPresentationColumn(formatter=my_func)

    where my_func is callable::

        my_func(value)
    """

    def __init__(self, **kwargs):
        """Initialize StringPresentationColumn. """
        self.formatter = kwargs.pop('formatter', str)
        super(StringPresentationColumn, self).__init__(**kwargs)

    def get_value(self, view, obj):
        """Extract value from the object for this column.

        Called in listing body.
        """
        val = str(obj)
        return self.formatter(val)


class ControlsColumn(Column):
    """Render View / Edit / Delete buttons."""

    def __init__(
            self,
            id: str='controls',
            name: t.Optional[str]='Actions',
            header_template: t.Optional[str]='crud/column_header_controls.html',
            body_template: t.Optional[str]='crud/column_body_controls.html'
    ):
        """Initialize ControlsColumn.

        :param id: ID of the column.
        :param name: Column name, to be displayed on the header of the table.
        :param header_template: Path to header template to be used.
        :param body_template: Path to body template to be used.
        """
        super(ControlsColumn, self).__init__(id=id, name=name, header_template=header_template, body_template=body_template)


class FriendlyTimeColumn(Column):
    """Print both accurate time and humanized relative time."""

    def __init__(
            self,
            id: str,
            name: str,
            navigate_view_name: t.Optional[str]=None,
            timezone: t.Optional[str]=None,
            header_template: t.Optional[str]=None,
            body_template: t.Optional[str]='crud/column_body_friendly_time.html'
    ):
        """Initialize FriendlyTimeColumn.

        :param id: Column id. Must match field id on the model
        :param name: Column name, to be displayed on the header of the table.
        :param navigate_view_name: If set, make this column clickable and navigates to the traversed name. Options are "show", "edit", "delete"
        :param timezone: Timezone to be used.
        :param header_template: Path to header template to be used.
        :param body_template: Path to body template to be used.
        """
        if timezone:
            self.timezone = timezone

        super(FriendlyTimeColumn, self).__init__(id=id, name=name, navigate_view_name=navigate_view_name, header_template=header_template, body_template=body_template)


class Table:
    """Describe table columns to a CRUD listing view."""

    def __init__(self, columns: t.Optional[t.List[Column]]=None):
        """Initialize Table.

        :param columns: List of columns to be used to render the list view.
        """
        self.columns = columns or []

    def get_columns(self) -> t.List[Column]:
        """Return columns.

        :return: List of columns to be used to render the list view.
        """
        return self.columns



"""Default user model implementations.

Define how User, Group, UserGroup and Activation models are in relationship together.

These models are picked up in :py:meth:`websauna.system.Initializer.configure_user_models`.
"""
# Pyramid
from zope.interface import implementer

# SQLAlchemy
import sqlalchemy as sa
from sqlalchemy.ext.declarative.base import _declarative_constructor

# Websauna
from websauna.system.user import usermixin
from websauna.system.user.interfaces import IGroup
from websauna.system.user.interfaces import IUser
from websauna.system.user.usermixin import ActivationMixin
from websauna.system.user.usermixin import UserGroupMixin


@implementer(IUser)
class User(usermixin.UserMixin):
    """The default user implementation for Websauna.

    This is a concrete implementation of SQLAlchemy model.
    """

    # In PSQL "user", the automatically generated table name, is a reserved word
    __tablename__ = "users"

    __init__ = _declarative_constructor

    #: Current user activation instance for reset password for sign up email verification
    activation_id = sa.Column(sa.Integer, sa.ForeignKey("user_activation.id"))

    #: SQLAlchemy relationship for above
    activation = sa.orm.relationship('Activation', backref='user')


@implementer(IGroup)
class Group(usermixin.GroupMixin):
    """The default group implementation for Websauna."""

    __tablename__ = "group"

    __init__ = _declarative_constructor

    users = sa.orm.relationship(
        'User',
        secondary="usergroup",
        passive_deletes=True,
        passive_updates=True,
        backref="groups",
    )

    def __str__(self):
        """Representation of a Group object."""
        return "Group #{id}: {name}".format(id=self.id, name=self.name)


class UserGroup(UserGroupMixin):
    """Map one user to one group."""

    __tablename__ = "usergroup"

    # Default constructor
    __init__ = _declarative_constructor

    user_id = sa.Column(sa.ForeignKey("users.id"))
    group_id = sa.Column(sa.ForeignKey("group.id"))


class Activation(ActivationMixin):
    """The default implementation of user email activation token."""

    __tablename__ = "user_activation"

    # Default constructor
    __init__ = _declarative_constructor



def test_frontpage(web_server, browser, ini_settings):
    """Open the front page as anonymous user."""
    b = browser
    b.visit(web_server)

    site_name = ini_settings["websauna.site_name"]
    element = b.find_by_css("h1#front-page-title")

    assert len(element) == 1
    assert element.text == site_name



"""Tests for checking database sanity checks functions correctly."""

import datetime

# SQLAlchemy
from sqlalchemy import Column
from sqlalchemy.ext.declarative import declarative_base

# Websauna
from websauna.system.model.columns import UTCDateTime

# pytest

import pytest


def test_UTCDateTime_restricts_timezone_to_utc():

    Base = declarative_base()

    class TestUTCModel(Base):
        """A sample SQLAlchemy model to demostrate db conflicts. """

        __tablename__ = "utc_datetime_test"

        date = Column(UTCDateTime(timezone=datetime.timezone.utc), primary_key=True)

    assert TestUTCModel.date.type.timezone is True

    with pytest.raises(ValueError):
        class TestUTCModel2(Base):
            """A sample SQLAlchemy model to demostrate db conflicts. """

            __tablename__ = "utc_datetime_test2"

            date = Column(UTCDateTime(timezone=object), primary_key=True)



"""Patched Google provider until we get upstream fixed.

https://github.com/peterhudec/authomatic/issues/153
"""

from authomatic import core
from authomatic.providers.oauth2 import OAuth2


class Google(OAuth2):
    """
    Google |oauth2| provider.

    * Dashboard: https://console.developers.google.com/project
    * Docs: https://developers.google.com/accounts/docs/OAuth2
    * API reference: https://developers.google.com/gdata/docs/directory
    * API explorer: https://developers.google.com/oauthplayground/

    Supported :class:`.User` properties:

    * email
    * first_name
    * gender
    * id
    * last_name
    * link
    * locale
    * name
    * picture

    Unsupported :class:`.User` properties:

    * birth_date
    * city
    * country
    * nickname
    * phone
    * postal_code
    * timezone
    * username

    .. |oauth2| replace:: oauth2
    .. note::

        To get the user info, you need to activate the **Google+ API**
        in the **APIs & auth >> APIs** section of the`Google Developers Console
        <https://console.developers.google.com/project>`__.

    """

    user_authorization_url = 'https://accounts.google.com/o/oauth2/auth'
    access_token_url = 'https://accounts.google.com/o/oauth2/token'
    user_info_url = 'https://www.googleapis.com/oauth2/v3/userinfo?alt=json'

    user_info_scope = ['profile',
                       'email']

    supported_user_attributes = core.SupportedUserAttributes(
        id=False,
        email=True,
        name=True,
        first_name=True,
        last_name=True,
        locale=True,
        picture=True,

        # No longer supported as of 2016-08
        link=False,
        gender=False,
    )

    def __init__(self, *args, **kwargs):
        super(Google, self).__init__(*args, **kwargs)

        # Handle special Google requirements to be able to refresh the access token.
        if self.offline:
            if 'access_type' not in self.user_authorization_params:
                # Google needs access_type=offline param in the user authorization request.
                self.user_authorization_params['access_type'] = 'offline'
            if 'approval_prompt' not in self.user_authorization_params:
                # And also approval_prompt=force.
                self.user_authorization_params['approval_prompt'] = 'force'

    @classmethod
    def _x_request_elements_filter(cls, request_type, request_elements,
                                   credentials):
        """
        Google doesn't accept client ID and secret to be at the same time in
        request parameters and in the basic authorization header in the
        access token request.
        """
        if request_type is cls.ACCESS_TOKEN_REQUEST_TYPE:
            params = request_elements[2]
            del params['client_id']
            del params['client_secret']
        return request_elements

    @staticmethod
    def _x_user_parser(user, data):

        user.email = data.get('email')
        user.name = data.get('name')
        user.first_name = data.get('given_name', '')
        user.last_name = data.get('family_name', '')
        user.locale = data.get('locale', '')
        user.picture = data.get('picture', '')

        # Special attribute if email has been verified
        user.email_verified = data.get("email_verified")
        user.hosted_domain = data.get("hd")
        return user

    def _x_scope_parser(self, scope):
        """
        Google has space-separated scopes
        """
        return ' '.join(scope)


# Needed for Authomatic magic
PROVIDER_ID_MAP = [Google]



# Standard Library
import datetime

# SQLAlchemy
from sqlalchemy import types


class UTCDateTimeTypeDecorator(types.TypeDecorator):
    """Automatically make datetimes to UTC when loaded from the database.

    http://stackoverflow.com/a/2528453/315168
    """

    impl = types.DateTime

    def process_bind_param(self, value, engine):
        if value is not None:
            return value.astimezone(datetime.timezone.utc)

    def process_result_value(self, value, engine):
        if value is not None:
            return datetime.datetime(value.year, value.month, value.day,
                            value.hour, value.minute, value.second,
                            value.microsecond, tzinfo=datetime.timezone.utc)



# Standard Library
import io
import logging
import os
from tempfile import NamedTemporaryFile

# Pyramid
import transaction
from pyramid import testing
from pyramid.threadlocal import get_current_registry

# Websauna
from websauna.system import get_init
from websauna.system.core.utils import get_secrets
from websauna.system.devop.backup import backup_site


# celery.utils.log module contains a big F"#€"#€)(!"(#! surprise which screws up all your logging, implicit calls to ensure_process_aware_logger()

Logger = logging.getLoggerClass()
Logger._process_aware = True


def test_backup(dbsession, ini_settings):
    """Execute backup script with having our settings content."""

    f = NamedTemporaryFile(delete=False)
    temp_fname = f.name
    f.close()

    ini_settings["websauna.backup_script"] = "websauna.system.tests:backup_script.bash"
    ini_settings["backup_test.filename"] = temp_fname

    # We have some scoping issues with the dbsession here, make sure we close transaction at the end of the test
    with transaction.manager:

        init = get_init(dict(__file__=ini_settings["_ini_file"]), ini_settings)
        init.run()

        testing.setUp(registry=init.config.registry)

        # Check we have faux AWS variable to export
        secrets = get_secrets(get_current_registry())
        assert "aws.access_key_id" in secrets

        try:

            # This will run the bash script above
            backup_site()

            # The result should be generated here
            assert os.path.exists(temp_fname)
            contents = io.open(temp_fname).read()

            # test-secrets.ini, AWS access key
            assert contents.strip() == "foo"
        finally:
            testing.tearDown()



"""User login and sign up handling views.


.. note ::

    Plan to move something more flexible in the future from these hardcoded, not very feature rich or flexible, views.

"""
# Standard Library
import logging

# Pyramid
import deform
from pyramid.httpexceptions import HTTPFound
from pyramid.httpexceptions import HTTPMethodNotAllowed
from pyramid.httpexceptions import HTTPNotFound
from pyramid.response import Response
from pyramid.settings import aslist
from pyramid.view import view_config

# Websauna
from websauna.system.core import messages
from websauna.system.core.route import get_config_route
from websauna.system.http import Request
from websauna.system.user.interfaces import AuthenticationFailure
from websauna.system.user.interfaces import CannotResetPasswordException
from websauna.system.user.interfaces import IForgotPasswordForm
from websauna.system.user.interfaces import IForgotPasswordSchema
from websauna.system.user.interfaces import ILoginForm
from websauna.system.user.interfaces import ILoginSchema
from websauna.system.user.interfaces import IRegisterForm
from websauna.system.user.interfaces import IRegisterSchema
from websauna.system.user.interfaces import IResetPasswordForm
from websauna.system.user.interfaces import IResetPasswordSchema
from websauna.system.user.utils import get_credential_activity_service
from websauna.system.user.utils import get_login_service
from websauna.system.user.utils import get_oauth_login_service
from websauna.system.user.utils import get_registration_service


logger = logging.getLogger(__name__)


@view_config(route_name='register', renderer='login/register.html')
def register(request: Request) -> dict:
    """Sign up view.

    :param request: Pyramid request.
    :return: Context to be used by the renderer.
    """
    settings = request.registry.settings

    schema = request.registry.getUtility(IRegisterSchema)
    schema = schema().bind(request=request)

    form_class = request.registry.getUtility(IRegisterForm)
    form = form_class(schema)

    social_logins = aslist(settings.get("websauna.social_logins", ""))

    if request.method == "POST":
        # If the request is a POST:
        controls = request.POST.items()
        try:
            captured = form.validate(controls)
        except deform.ValidationFailure as e:
            return {'form': e.render(), 'errors': e.error.children, 'social_logins': social_logins}

        # With the form validated, we know email and username are unique.
        del captured['csrf_token']

        registration_service = get_registration_service(request)
        return registration_service.sign_up(user_data=captured)

    return {'form': form.render(), 'social_logins': social_logins}


@view_config(route_name='activate')
def activate(request: Request) -> dict:
    """View to activate user after clicking email link.

    :param request: Pyramid request.
    :return: Context to be used by the renderer.
    """
    code = request.matchdict.get('code', None)
    registration_service = get_registration_service(request)
    return registration_service.activate_by_email(code)


@view_config(route_name='login', renderer='login/login.html')
def login(request: Request) -> [HTTPFound, dict]:
    """Default login view implementation.

    :param request: Pyramid request.
    :return: Context to be used by the renderer or a HTTPFound redirect if user is already logged in.
    """
    login_redirect_view = get_config_route(request, "websauna.login_redirect")

    # Create login form schema
    schema = request.registry.getUtility(ILoginSchema)
    schema = schema().bind(request=request)

    # Create login form
    form = request.registry.getUtility(ILoginForm)
    form = form(schema)
    settings = request.registry.settings

    social_logins = aslist(settings.get("websauna.social_logins", ""))

    # Process form
    if request.method == "POST":
        try:
            controls = request.POST.items()
            captured = form.validate(controls)
        except deform.ValidationFailure as e:
            return {
                'form': e.render(),
                'errors': e.error.children
            }

        username = captured['username']
        password = captured['password']
        login_service = get_login_service(request)

        try:
            return login_service.authenticate_credentials(username, password, login_source="login_form")
        except AuthenticationFailure as e:
            # Tell user they cannot login at the moment
            messages.add(request, msg=str(e), msg_id="msg-authentication-failure", kind="error")

            return {
                'form': form.render(appstruct=captured),
                'errors': [e],
                "social_logins": social_logins
            }
    else:
        # HTTP get, display login form
        if request.user:
            # Already logged in
            return HTTPFound(location=login_redirect_view)

        # Display login form
        return {'form': form.render(), "social_logins": social_logins}


@view_config(route_name='registration_complete', renderer='login/registration_complete.html')
def registration_complete(request: Request) -> dict:
    """After activation initial login screen.

    :param request: Pyramid request.
    :return: Context to be used by the renderer.
    """
    # Create login form schema
    schema = request.registry.getUtility(ILoginSchema)
    schema = schema().bind(request=request)

    # Create login form
    form = request.registry.getUtility(ILoginForm)
    form = form(schema)

    # Make sure we POST to right endpoint
    form.action = request.route_url('login')

    return {"form": form.render()}


@view_config(route_name='login_social')
def login_social(request: Request) -> dict:
    """Login using OAuth and any of the social providers.

    :param request: Pyramid request.
    :return: Context to be used by the renderer.
    """
    # Get the internal provider name URL variable.
    provider_name = request.matchdict.get('provider_name')
    oauth_login_service = get_oauth_login_service(request)
    assert oauth_login_service, "OAuth not configured for {}".format(provider_name)
    return oauth_login_service.handle_request(provider_name)


@view_config(permission='authenticated', route_name='logout')
def logout(request: Request) -> Response:
    """Logout view.

    :param request: Pyramid request.
    :return: Context to be used by the renderer.
    """
    if request.method != "POST":
        # No GET / CSRF logouts
        return HTTPMethodNotAllowed()
    login_service = get_login_service(request)
    return login_service.logout()


@view_config(route_name='forgot_password', renderer='login/forgot_password.html')
def forgot_password(request: Request) -> dict:
    """Forgot password screen.

    :param request: Pyramid request.
    :return: Context to be used by the renderer.
    """
    schema = request.registry.getUtility(IForgotPasswordSchema)
    schema = schema().bind(request=request)

    form = request.registry.getUtility(IForgotPasswordForm)
    form = form(schema)

    if request.method == 'POST':
        # From here on, we know it's a POST. Let's validate the form
        controls = request.POST.items()
        try:
            captured = form.validate(controls)
        except deform.ValidationFailure as e:
            # This catches if the email does not exist, too.
            return {'form': e.render(), 'errors': e.error.children}

        credential_activity_service = get_credential_activity_service(request)
        # Process valid form
        email = captured["email"]
        try:
            return credential_activity_service.create_forgot_password_request(email)
        except CannotResetPasswordException as e:
            messages.add(request, msg=str(e), msg_id="msg-cannot-reset-password", kind="error")
            return {'form': form.render()}
    return {'form': form.render()}


@view_config(route_name='reset_password', renderer='login/reset_password.html')
def reset_password(request: Request) -> dict:
    """Reset password view.

    User arrives on the page and enters the new password.

    :param request: Pyramid request.
    :return: Context to be used by the renderer.
    """
    schema = request.registry.getUtility(IResetPasswordSchema)
    schema = schema().bind(request=request)

    form = request.registry.getUtility(IResetPasswordForm)
    form = form(schema)

    code = request.matchdict.get('code', None)
    credential_activity_service = get_credential_activity_service(request)
    user = credential_activity_service.get_user_for_password_reset_token(code)
    if not user:
        raise HTTPNotFound("Invalid password reset code")
    if request.method == 'POST':
        try:
            controls = request.POST.items()
            captured = form.validate(controls)
        except deform.ValidationFailure as e:
            return {'form': e.render(), 'errors': e.error.children}
        password = captured['password']
        return credential_activity_service.reset_password(code, password)
    # Question screen
    return {
        'form': form.render(
            appstruct=dict(
                user=user.friendly_name
            )
        )
    }



"""Scaffold test utility functions."""
# Standard Library
import os
import subprocess
import sys
import time
import typing as t  # noQA
from contextlib import closing
from contextlib import contextmanager
from tempfile import mkdtemp

# SQLAlchemy
import psycopg2

import pytest
from cookiecutter.main import cookiecutter

from ..utils import execute_command
from ..utils import print_subprocess_fail


PYTHON_INTERPRETER = "python{}.{}".format(sys.version_info.major, sys.version_info.minor)


def execute_venv_command(cmdline, folder, timeout=15.0, wait_and_see=None, assert_exit=0, cd_folder=None):
    """Run a command in a specific folder using virtualenv created there.

    Assume virtualenv is under ``venv`` folder.

    :param wait_and_see: Wait this many seconds to see if app starts up.
    :param assert_exit: Assume exit code is this
    :param cd_folder: cd to this folder before executing the command (relative to folder)
    :return: tuple (exit code, stdout, stderr)
    """

    assert os.path.exists(os.path.join(folder, "venv", "bin", "activate"))

    if type(cmdline) == list:
        cmdline = " ".join(cmdline)

    if cd_folder is not None:
        cd_cmd = "cd {} && ".format(cd_folder)
    else:
        cd_cmd = ""

    cmdline = ". {}/venv/bin/activate ; {} {}".format(folder, cd_cmd, cmdline)

    # print("Executing {} in {}".format(cmdline, folder))

    worker = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=folder, shell=True)

    if wait_and_see is not None:
        time.sleep(wait_and_see)
        worker.poll()

        if worker.returncode is not None:
            # Return code is set if the worker dies within the timeout
            print_subprocess_fail(worker, cmdline)
            raise AssertionError("could not start server like app: {}".format(cmdline))

        worker.kill()
        return 0
    else:
        try:
            worker.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            print_subprocess_fail(worker, cmdline)
            raise

    if worker.returncode != assert_exit:
        print_subprocess_fail(worker, cmdline)
        raise AssertionError("venv command did not properly exit: {} in {}. Got exit code {}, assumed {}".format(cmdline, folder, worker.returncode, assert_exit))

    return (worker.returncode, worker.stdout.read().decode("utf-8"), worker.stderr.read().decode("utf-8"))


def preload_wheelhouse(folder: str):
    """Speed up tests by loading Python packages from primed cache.

    Use ``create_wheelhouse.bash`` to prime the cache.

    :param folder: Temporary virtualenv installation
    """
    cache_folder = os.getcwd()
    wheelhouse_folder = os.path.join(
        cache_folder,
        "wheelhouse", "python{major}.{minor}".format(
            major=sys.version_info.major,
            minor=sys.version_info.minor
        )
    )
    if os.path.exists(wheelhouse_folder):
        execute_venv_command(
            "pip install {wheelhouse_folder}/*".format(wheelhouse_folder=wheelhouse_folder),
            folder,
            timeout=3 * 60
        )
    else:
        print("No preloaded Python package cache found")


def create_psq_db(request, dbname, dsn=''):
    """py.test fixture to createdb and destroy postgresql database on demand."""
    if not dsn:
        dsn = "dbname=postgres"
    with closing(psycopg2.connect(dsn)) as conn:
        conn.autocommit = True
        with closing(conn.cursor()) as cursor:
            cursor.execute("SELECT COUNT(*) FROM pg_database WHERE datname='{}'".format(dbname))

            if cursor.fetchone()[0] == 1:
                # Prior interrupted test run
                cursor.execute('DROP DATABASE ' + dbname)

            cursor.execute('CREATE DATABASE ' + dbname)

    def teardown():
        with closing(psycopg2.connect(dsn)) as conn:
            conn.autocommit = True
            with closing(conn.cursor()) as cursor:

                # http://blog.gahooa.com/2010/11/03/how-to-force-drop-a-postgresql-database-by-killing-off-connection-processes/
                cursor.execute("SELECT pg_terminate_backend(pid) from pg_stat_activity where datname='{}';".format(dbname))
                conn.commit()
                cursor.execute("SELECT COUNT(*) FROM pg_database WHERE datname='{}'".format(dbname))
                if cursor.fetchone()[0] == 1:
                    cursor.execute('DROP DATABASE ' + dbname)

    request.addfinalizer(teardown)


@contextmanager
def replace_file(path: str, content: str):
    """A context manager to temporary swap the content of a file.

    :param path: Path to the file
    :param content: New content as a text
    """
    backup = open(path, "rt").read()
    open(path, "wt").write(content)

    try:
        yield None
    finally:
        open(path, "wt").write(backup)


@contextmanager
def insert_content_after_line(path: str, content: str, marker: str):
    """Add piece to text to a text file after a line having a marker string on it."""
    backup = open(path, "rt").read()
    try:
        # Replaces stdout
        out = open(path, "wt")
        for line in backup.split("\n"):
            if marker in line:
                print(content, file=out)
            print(line, file=out)
        out.close()
        yield None
    finally:
        open(path, "wt").write(backup)


@pytest.fixture(scope='session')
def cookiecutter_config(tmpdir_factory) -> str:
    """py.test fixture to generate a tmp config file for cookiecutter.

    :return: Path to cookiecutter config file.
   """
    user_dir = tmpdir_factory.mktemp('user_dir')
    cookiecutters_dir = user_dir.mkdir('cookiecutters')
    replay_dir = user_dir.mkdir('cookiecutter_replay')
    USER_CONFIG = '''cookiecutters_dir: "{cookiecutters_dir}"\nreplay_dir: "{replay_dir}"'''
    config_text = USER_CONFIG.format(
        cookiecutters_dir=cookiecutters_dir,
        replay_dir=replay_dir,
    )
    config_file = user_dir.join('config')

    config_file.write_text(config_text, encoding='utf8')
    return str(config_file)


@pytest.fixture(scope='session')
def app_scaffold(request, cookiecutter_config) -> str:
    """py.test fixture to create app scaffold.

    Create application and virtualenv for it. Run setup.py.

    :return: Path to a temporary folder. In this folder there is `venv` folder and `myapp` folder.
   """

    folder = mkdtemp(prefix="websauna_test_")

    websauna_folder = os.getcwd()
    execute_command([PYTHON_INTERPRETER, "-m", "venv", "venv"], folder, timeout=30)

    # Don't try to push to get a working pip because IT "#€!"#€"#€ DOESNT'T JUST WORK
    # Instead work around any issues caused by missing pip in tests themselves
    #
    # venv fails to install pip under .tox virtualenv due to some obscure bug
    # This broken Python venv stuff drives me crazy... make sure we get  pip
    # pip = os.path.join(folder, "venv", "bin", "pip")
    # if not os.path.exists(pip):
    # Use internal get-pip script to fix broken venv where ensurepip did no give us our shit, because we can't rely on system pip to get this right either
    #   print("The current version of venv/ensurepip modules are broken, fixing problem internally")
    #   get_pip = os.path.join(os.path.dirname(__file__), "get-pip.py")
    #   assert os.path.exists(get_pip)
    #   execute_venv_command("python {} --prefix {}/venv --ignore-installed pip".format(get_pip, folder), folder, timeout=5*60)

    # assert os.path.exists(pip), "Pip not installed: {}".format(pip)

    # PIP cannot handle pip -install .[test]
    # On some systems, the default PIP is too old and it doesn't seem to allow upgrade through wheelhouse
    execute_venv_command("pip install -U pip", folder, timeout=5 * 60)

    # Install cached PyPi packages
    preload_wheelhouse(folder)

    # Install websauna
    execute_venv_command("cd {websauna_folder} ; pip install -e .[notebook,utils]".format(websauna_folder=websauna_folder), folder, timeout=5 * 60)

    # Create Websauna app, using cookiecutter, from template cookiecutter-websauna-app
    extra_context = {
        'full_name': 'Websauna Team',
        'email': 'developers@websauna.org',
        'company': 'Websauna',
        'github_username': 'websauna',
        'project_name': 'Websauna: News portal',
        'project_short_description': 'Websauna news portal application.',
        'tags': 'python package websauna pyramid',
        'repo_name': 'my.app',
        'namespace': 'my',
        'package_name': 'app',
        'release_date': 'today',
        'year': '2017',
        'version': '1.0.0a1',
        'create_virtualenv': 'No'
    }
    template = 'https://github.com/websauna/cookiecutter-websauna-app/archive/master.zip'
    cookiecutter(
        template,
        no_input=True,
        extra_context=extra_context,
        output_dir=folder,
        config_file=cookiecutter_config
    )

    # Install the package created by cookiecutter template
    content_folder = os.path.join(folder, extra_context["repo_name"])
    execute_venv_command("pip install -e {0}".format(content_folder), folder, timeout=5 * 60)

    def teardown():
        # Clean any processes who still think they want to stick around. Namely: ws-shell doesn't die

        # This kills all processes referring to the temporary folder
        subprocess.call("pkill -SIGKILL -f {}".format(folder), shell=True)

    request.addfinalizer(teardown)
    return folder


def start_ws_pserve(cmdline: str, cwd: str, wait_and_see: float=5.0):
    """Simulate starting ws-pserve command from the command line inside the virtualenv.

    :param cmdline: Command line to run ws-pserve
    :param cwd: Set current workind directory
    :param wait_and_see: Seconds to see if the server comes up
    :return:
    """

    # Clean up all prior processes
    import psutil
    import signal

    # http://stackoverflow.com/a/20691431/315168
    for proc in psutil.process_iter():
        try:
            for conns in proc.connections(kind='inet'):
                if conns.laddr[1] == 6543:
                    print("Killing a proc blocking the port", proc)
                    proc.send_signal(signal.SIGKILL)
                    time.sleep(0.5)
                    continue
        except (psutil.AccessDenied, psutil.NoSuchProcess):
            pass

    # Run ws-pserve inside the virtualenc
    cmdline = ". {}/venv/bin/activate && ".format(cwd) + cmdline

    worker = subprocess.Popen(cmdline, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)

    time.sleep(wait_and_see)
    worker.poll()

    if worker.returncode is not None:
        # Return code is set if the worker dies within the timeout
        print_subprocess_fail(worker, cmdline)
        raise AssertionError("Could not ws-pserve: {}".format(cmdline))

    return worker



# Pyramid
from pyramid.config import Configurator

import pytest
from webtest import TestApp


def setup_wsgi():
    configurator = Configurator()

    configurator.include("pyramid_jinja2")
    configurator.add_jinja2_renderer('.html')
    configurator.add_jinja2_search_path('websauna.system.tests:templates/viewconfig', name='.html')

    configurator.add_route("parent_hello", "/parent_hello")
    configurator.add_route("child_hello", "/child_hello")

    from websauna.system.tests.viewconfig import testmodule
    configurator.set_root_factory(testmodule.Root)
    configurator.scan(testmodule)

    wsgi = TestApp(configurator.make_wsgi_app())
    return wsgi


def test_override_context():
    """Subclasses may override the context for which the view belongs.."""
    wsgi = setup_wsgi()

    resp = wsgi.get("/parent/edit", status=200)
    assert resp.text == '200 OK\n\n\n\n\nEditing: parent\n\n'

    resp = wsgi.get("/child/edit", status=200)
    assert resp.text == '200 OK\n\n\n\n\nEditing: child\n\n'


def test_child_own_view():
    """Subclasses can defined their own view configs."""
    wsgi = setup_wsgi()

    resp = wsgi.get("/child2/show", status=200)
    assert resp.text == '200 OK\n\n\n\n\nShowing: child2\n\n'


def test_grand_child_view():
    """view_overrides work correctly on the second level of inheritance."""
    wsgi = setup_wsgi()

    resp = wsgi.get("/grand_child/edit", status=200)
    assert resp.text == '200 OK\n\n\n\n\nEditing: grand_child\n\n'


def test_grand_grand_child_view():
    """view_overrides work correctly on the third level of inheritance with two nested @view_overrides."""
    wsgi = setup_wsgi()

    resp = wsgi.get("/grand_grand_child/edit", status=200)
    assert resp.text == '200 OK\n\n\n\n\nEditing: grand_grand_child\n\n'


def test_invalid_override():
    """Trying to override non-view class causes an error.."""
    Configurator()
    with pytest.raises(RuntimeError):
        from websauna.system.tests.viewconfig import brokentestmodule  # noqa


def test_rendered_override():
    """We can override renderer and route name for dispatched URLs."""
    wsgi = setup_wsgi()

    resp = wsgi.get("/parent_hello", status=200)
    assert resp.text.strip() == 'Hello Parent: ParentRouteView'

    resp = wsgi.get("/child_hello", status=200)
    assert resp.text.strip() == 'Hello Child: ChildRouteView'



"""Test view throttlign facilities."""
# Pyramid
from pyramid.httpexceptions import HTTPOk

import pytest
from webtest import TestApp

# Websauna
import websauna
from websauna.system.core.redis import get_redis  # noQA
from websauna.system.form.throttle import clear_throttle
from websauna.system.form.throttle import throttled_view


def throttle_sample(request):
    return HTTPOk()


@pytest.fixture(scope="module")
def throttle_app(request, paster_config):
    '''Custom WSGI app with permission test views enabled.'''

    class Initializer(websauna.system.Initializer):

        def configure_views(self):
            self.config.add_route("throttle_sample", "/")
            self.config.add_view(throttle_sample, route_name="throttle_sample", decorator=throttled_view(limit=1))

        def configure_csrf(self):
            """Disable CSRF for this test run for making testing simpler."""
            self.config.set_default_csrf_options(require_csrf=False)

    global_config, app_settings = paster_config
    init = Initializer(global_config, app_settings)
    init.run()
    app = TestApp(init.make_wsgi_app())
    app.init = init
    return app


def test_throttle(throttle_app: TestApp, test_request):
    """Throttling should give us 429."""

    app = throttle_app

    # clear counter from previous test run
    clear_throttle(test_request, "throttle_sample")

    app.get("/", status=200)

    # We exceeded the limit of 1 request per hour
    app.get("/", status=429)

    # Let's clear the counter
    clear_throttle(test_request, "throttle_sample")

    app.get("/", status=200)



"""ws-sanity-check script.

Execute a sanity check on the configuration.
"""
# Standard Library
import sys
import typing as t

# Websauna
from websauna.system import SanityCheckFailed
from websauna.system.devop.cmdline import init_websauna
from websauna.system.devop.scripts import FAIL_MSG
from websauna.system.devop.scripts import feedback_and_exit
from websauna.system.devop.scripts import get_config_uri
from websauna.system.devop.scripts import usage_message


SUCCESS_MSG = """Sanity check passed"""

NOT_FOUND_MSG = """
Sanity check failed

Reason:
Config {config_uri} was not found.
"""


def main(argv: t.List[str]=sys.argv):
    """Execute a sanity check on the configuration.

    :param argv: Command line arguments, second one needs to be the uri to a configuration file.
    :raises sys.SystemExit:
    """
    if len(argv) < 2:
        usage_message(argv)

    config_uri = get_config_uri(argv)

    try:
        init_websauna(config_uri, sanity_check=True)
    except SanityCheckFailed as exc:
        feedback_and_exit(FAIL_MSG.format(exception=str(exc)), 10)
    except FileNotFoundError as exc:
        feedback_and_exit(NOT_FOUND_MSG.format(config_uri=config_uri), 10)

    feedback_and_exit(SUCCESS_MSG, 0)



"""Traversing core logic."""
# Pyramid
from pyramid.interfaces import ILocation
from zope.interface import implementer


@implementer(ILocation)
class Resource:
    """Traversable resource in a nested tree hierarchy with basic breadcrumbs support.

    All traverable context classes should inherit from this class. Note that this is not a strict requirement, as often anything implementing :py:class:`pyramid.interfaces.ILocation` and ``get_title()`` will work.

    For more information see :ref:`Traversal <traversal>`.

    ..  _traversal:
    """

    # TODO: Cannot annotate request as it breaks sphinx-autodoc-typehints, sphinx-autodoc-typehints==1.1.0, when doing make html
    def __init__(self, request):

        #: Pointer to the parent object in traverse hierarchy. This is none until make_lineage is called.
        self.__parent__ = None

        #: The id of this resource as its appear in URL and traversing path
        self.__name__ = None

        self.request = request

    def get_title(self) -> str:
        """Return human-readable title of this resource.

        This is viewed in admin breadcrumbs path, etc.
        """
        title = getattr(self, "title", None)
        if title:
            return title

        raise NotImplementedError("get_title() implementation missing for {}".format(self))

    @classmethod
    def make_lineage(self, parent, child, name, allow_new_parent=False) -> "Resource":
        """Set traversing pointers between the child and the parent resources.

        Builds __parent__ and __name__ pointer and sets it on the child resource.

        * If lineage relationship is not lazy and the referenced children is stored in the parent, the lineage must be set when the child is put into parent container.

        * If lineage relationship is lazy and child resource is constructed upon lookup in ``__item__``, the lineage is constructed before the child is returned.

        :param parent: Parent resource who children is become part to

        :param child: Child resource mutated in place

        :param name: Id of the child resource as it will appear in the URL traversing path

        :param allow_new_parent: If the child has alraedy a parent assigned, allow override the parent... or basically move an existing resource. You don't usually want this for in-memory resource and this is for catching bugs.

        :return: The mutated child resource
        """

        assert child
        assert parent
        assert name

        if not allow_new_parent:
            # Catch bugs when you try to double lineage a persistnet parent -> child relationship
            assert not getattr(child, "__parent__", None), "Tried to double init lineage for {} -> {}, previous parent was {}".format(parent, child, child.__parent__)

        child.__parent__ = parent
        child.__name__ = name
        return child



# Standard Library
import typing as t

# Pyramid
from pyramid.interfaces import ILocation
from zope.interface import Interface


class IRoot(Interface):
    """Market interface for the root object.

    Used e.g. breadcrumbs and traversing are we in root tests.
    """


class ISecrets(Interface):
    """Utility marker interface which gives us our secrets.

    Secrets is a dictionary which hold sensitive deployment data.
    """


class IContainer(ILocation):
    """Marker interface telling that Resource class is iterable.

    Database loading example implementation:

    .. code-block:: python

        @implementer(IContainer)
        class AssetFolder(Resource):

            def __getitem__(self, slug: str) -> AssetDescription:

                for asset in self.request.dbsession.query(Asset).filter(Asset.network_id==self.get_network().id):
                    if asset.slug == slug:
                        return self.get_description(asset)

                raise KeyError()

            def items(self):
                for asset_resource in self.get_public_assets():
                    yield asset_resource.__name__, asset_resource

            def get_public_assets(self) -> Iterable[AssetDescription]:
                network = self.get_network()
                dbsession = self.request.dbsession
                for asset in dbsession.query(Asset).filter_by(network=network, state=AssetState.public).order_by(Asset.name.asc()):
                    yield self.get_description(asset)

    Static child example implementation:

    .. code-block:: python

        @implementer(IContainer)
        class NetworkDescription(Resource):

            def items(self):
                yield "assets", self.asset_folder

            def __getitem__(self, item):
                if item == "assets":
                    return self.asset_folder
                raise KeyError()

    For more example implementation see :py:mod:`websauna.tests.sitemapsamples`.

    TODO: Not sure if items() is the best way to do the child discovery. All ideas accepted.
    """

    def items() -> t.Iterable[t.Tuple[str, ILocation]]:
        """Return children in this container as (id, Resource instance) tuples.

        This usually dynamically populates from the database when this is called and there is no caching. The result is iterable only once.

        :return: Iterable (URL id, child object). Child objects can be any Python objects with `__parent__` pointer set as described by a generic interface :py:class:`pyramid.interfaces.ILocation`.
        """



"""Sample application demostrating SQL free functionality."""

# Websauna
import websauna.system


class RedisUser:
    pass


class RedisUserRegistry:
    pass


class Initializer(websauna.system.Initializer):
    """An initialization configuration used for starting myapp.

    Override parent class methods to customize application behavior.
    """

    def configure_templates(self):
        """Include our package templates folder in Jinja 2 configuration."""
        super(Initializer, self).configure_templates()

    def configure_views(self):
        """Configure views for your application.

        Let the config scanner to pick ``@simple_route`` definitions from scanned modules. Alternative you can call ``config.add_route()`` and ``config.add_view()`` here.
        """
        # We override this method, so that we route home to our home screen, not Websauna default one
        from . import views
        self.config.scan(views)

    def configure_models(self):
        """Register the models of this application."""
        from . import models
        self.config.scan(models)

    def configure_model_admins(self):
        """Register the models of this application."""

        # Call parent which registers user and group admins
        super(Initializer, self).configure_model_admins()

        # Scan our admins
        from . import admins
        self.config.scan(admins)

    def run(self):
        super(Initializer, self).run()



"""ws-collect-static script.

Read through all configured static views and compile their assets to ``collected-static`` folder.
"""
# Standard Library
import sys
import typing as t

# Websauna
from websauna.system.devop.cmdline import init_websauna
from websauna.system.devop.scripts import feedback_and_exit
from websauna.system.devop.scripts import get_config_uri
from websauna.system.devop.scripts import usage_message


def main(argv: t.List[str]=sys.argv):
    """Read through all configured static views and compile their assets to ``collected-static`` folder.

    :param argv: Command line arguments, second one needs to be the uri to a configuration file.
    :raises sys.SystemExit:
    """
    if len(argv) < 2:
        usage_message(argv)

    config_uri = get_config_uri(argv)
    request = init_websauna(config_uri, sanity_check=False)
    request.registry.static_asset_policy.collect_static()
    feedback_and_exit('ws-collect-static: Collected all static assets', 0, True)



"""Default admin root implementation."""

# Pyramid
from pyramid.security import Allow
from pyramid.security import Deny
from pyramid.security import Everyone
from zope.interface import implementer

# Websauna
from websauna.system.admin import menu
from websauna.system.admin.events import AdminConstruction
from websauna.system.admin.interfaces import IAdmin
from websauna.system.core.root import Root
from websauna.system.core.traversal import Resource


@implementer(IAdmin)
class Admin(Resource):
    """Admin interface main object.

    Presents /admin part of the URL. Manages model admin registrations and discovery. Provides helper functions to map SQLAlchemy objects to their admin URLs.

    ``Admin`` declares two default menu systems which can be used to register application specific and third party add on entries

    * ``Admin.get_quick_menu()`` returns a vertical menu which is visible in the main site navigation

    * ``Admin.get_admin_menu()`` returns a horizontal menu which is visible after entering the admin UI

    This class is instiated for each request and does not have global state.
    """

    #: Default permissions of who can add, read and write things in admin
    __acl__ = [
        # Declare admin rights
        (Allow, 'group:admin', 'add'),
        (Allow, 'group:admin', 'view'),
        (Allow, 'group:admin', 'edit'),
        (Allow, 'group:admin', 'delete'),
        (Allow, 'superuser:superuser', 'shell'),

        # Disable access to public users
        (Deny, Everyone, 'view'),  # Declared in websauna.system.core.root.Root
    ]

    def __init__(self, request):
        super(Admin, self).__init__(request)

        # Current add_route() view config sets Admin instance as request.root when traversing inside admin.
        # Assume this admin instance lives directly under the root
        self.__parent__ = Root(request)
        self.__name__ = "admin"

        self.admin_menu_entry = None
        self.quick_menu_entry = None

        # Registered child resources, usually put in AdminConstruct event
        self.children = {}

        self.construct()

    def get_title(self):
        return "Admin"

    def construct(self):
        """Call all admin contributors and let them register parts to this admin."""

        self.construct_default_menu()

        # Call all plugins to register themselves
        self.request.registry.notify(AdminConstruction(self))

    def construct_default_menu(self):
        """Setup admin main menu."""

        self.admin_menu_entry = menu.NavbarEntry("admin-menu-navbar", label=None, submenu=menu.Menu(), css_class="navbar-admin")
        self.quick_menu_entry = menu.RouteEntry("admin-menu-quick", "Admin", "admin_home", icon="fa-wrench", submenu=menu.Menu())

        home = menu.RouteEntry("admin-quick-menu-home", "Dashboard", "admin_home", icon="fa-wrench")
        self.quick_menu_entry.submenu.add_entry(home)

        home = menu.RouteEntry("admin-menu-home", "Dashboard", "admin_home", icon="fa-wrench")
        self.admin_menu_entry.submenu.add_entry(home)

        data = menu.RouteEntry("admin-menu-data", "Data", "admin_home", submenu=menu.Menu(), icon="fa-list")

        self.admin_menu_entry.submenu.add_entry(data)

    def get_quick_menu_entry(self) -> menu.Entry:
        """Return Admin root menu."""
        return self.quick_menu_entry

    def get_quick_menu(self) -> menu.Menu:
        return self.quick_menu_entry.submenu

    def get_admin_menu_entry(self) -> menu.Entry:
        return self.admin_menu_entry

    def get_admin_menu(self) -> menu.Menu:
        return self.admin_menu_entry.submenu

    def __getitem__(self, name):
        """Traverse to individual model admins by the model name."""
        child = self.children[name]
        return child



# Pyramid
import transaction

# Websauna
from websauna.system.user.models import User


def test_user_subsystem(init, dbsession):
    """Load the default user models and see we create correponding tables right."""

    with transaction.manager:
        u = User(email="example@example.com")
        dbsession.add(u)

    with transaction.manager:
        u = dbsession.query(User).get(1)
        assert u.email == "example@example.com"



"""ws-create-table script.

Print out sql statements needed to construct all the currently configured models.
"""
# Standard Library
import sys
import typing as t

# SQLAlchemy
from sqlalchemy.sql.ddl import CreateTable

# Websauna
from websauna.system.devop.cmdline import init_websauna
from websauna.system.devop.scripts import get_config_uri
from websauna.system.devop.scripts import usage_message
from websauna.system.model.meta import Base


def main(argv: t.List[str]=sys.argv):
    """Print out sql statements needed to construct currently configured models.

    :param argv: Command line arguments, second one needs to be the uri to a configuration file.
    :raises sys.SystemExit:
    """
    if len(argv) < 2:
        usage_message(argv)

    config_uri = get_config_uri(argv)
    request = init_websauna(config_uri)
    engine = request.dbsession.get_bind()

    for name, cls in Base._decl_class_registry.items():

        if name == "_sa_module_registry":
            continue

        print(CreateTable(cls.__table__, bind=engine))


if __name__ == "__main__":
    main()



# Pyramid
import transaction

import pytest
import requests
from flaky import flaky
from splinter.driver import DriverAPI

# Websauna
from websauna.system.user.models import User
from websauna.system.tests.utils import create_logged_in_user
from websauna.system.tests.utils import create_user
from websauna.system.tests.utils import logout
from websauna.utils.slug import uuid_to_slug


@pytest.fixture
def victim_browser(request, browser_instance_getter):
    """py.test fixture to open up another web browser for password reset victim."""
    return browser_instance_getter(request, victim_browser)


def test_view_user_details(browser, web_server, init, dbsession):
    """See that we can view the details of the user in a browser."""

    b = browser

    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)

    b.find_by_css("#nav-admin").click()

    b.find_by_css("#latest-user-shortcut").click()

    # TODO: Use CSS selector
    assert b.is_text_present("example@example.com")

    with transaction.manager:
        # Check that we show the user uuid slug on the page correctly
        u = dbsession.query(User).first()
        assert b.is_text_present(uuid_to_slug(u.uuid))


def test_add_user(browser, web_server, init, dbsession):
    """See that we can add new users."""

    b = browser

    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)

    b.find_by_css("#nav-admin").click()

    b.find_by_css("#btn-panel-add-user").click()

    # b.fill("username", "test2")
    b.fill("email", "test2@example.com")
    b.fill("password", "secret")
    b.fill("password-confirm", "secret")
    b.find_by_name("add").click()

    # TODO: Convert to CSS based test
    assert b.is_element_present_by_css("#msg-item-added")

    logout(web_server, b)

    b.visit(web_server + "/login")
    b.fill("username", "test2@example.com")
    b.fill("password", "secret")
    b.find_by_name("login_email").click()

    assert b.is_element_present_by_css("#msg-you-are-logged-in")


def test_add_user_password_mismatch(browser, web_server, init, dbsession):
    """Make sure new user is not created on password mismatch."""

    b = browser

    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)

    b.find_by_css("#nav-admin").click()

    b.find_by_css("#btn-panel-add-user").click()

    # b.fill("username", "test2")
    b.fill("email", "test2@example.com")
    b.fill("password", "secret")
    b.fill("password-confirm", "faied")
    b.find_by_name("add").click()

    # TODO: Convert to CSS based test
    assert b.is_text_present("Password did not match confirm")


def test_add_user_existing_email(browser, web_server, init, dbsession):
    """Add a user but there already exists one with the same email."""

    with transaction.manager:
        create_user(dbsession, init.config.registry, email="test2@example.com")

    b = browser

    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)

    b.find_by_css("#nav-admin").click()

    b.find_by_css("#btn-panel-add-user").click()

    # b.fill("username", "test2")
    b.fill("email", "test2@example.com")
    b.fill("password", "secret")
    b.fill("password-confirm", "secret")
    b.find_by_name("add").click()

    assert b.is_element_present_by_css("#error-deformField1")  # Email address already taken


def test_add_user_with_group(browser, web_server, init, dbsession):
    """Add a user and directly assign a group."""
    b = browser

    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)

    b.find_by_css("#nav-admin").click()

    b.find_by_css("#btn-panel-add-user").click()

    # b.fill("username", "test2")
    b.fill("email", "test2@example.com")
    b.fill("password", "secret")
    b.fill("password-confirm", "secret")

    # TODO: Make sure checkbox widget gets proper css classes
    b.find_by_css("input[type='checkbox']")[0].click()
    b.find_by_name("add").click()

    assert b.is_element_present_by_css("#msg-item-added")

    with transaction.manager:
        u = dbsession.query(User).get(2)
        assert len(u.groups) == 1


@flaky
def test_set_password(browser, victim_browser, web_server, init, dbsession):
    """See that we can reset the user password.

    1. Have admin user, normal user

    2. See that when the admin resets the user password all the user sessions are dropped (security feature)
    """

    b = browser
    b2 = victim_browser

    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)
    create_logged_in_user(dbsession, init.config.registry, web_server, b2, email="victim@example.com", password="secret")

    b.find_by_css("#nav-admin").click()
    b.find_by_css("#latest-user-shortcut").click()
    b.find_by_css("#btn-crud-set-password").click()

    b.fill("password", "new-secret")
    b.fill("password-confirm", "new-secret")
    b.find_by_name("save").click()

    assert b.is_element_present_by_css("#msg-password-changed")

    # Victim browser should have now logged out
    b2.visit(web_server)
    assert b2.is_element_present_by_css("#msg-session-invalidated")
    assert b2.is_element_present_by_css("#nav-sign-in")

    # See that we can log in with the new password
    b2.visit(web_server + "/login")
    b2.fill("username", "victim@example.com")
    b2.fill("password", "new-secret")
    b2.find_by_name("login_email").click()

    assert b2.is_element_present_by_css("#msg-you-are-logged-in")


def test_set_email(browser, victim_browser, web_server, init, dbsession):
    """Setting email resets user session and user can log in again."""

    b = browser
    b2 = victim_browser

    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)
    create_logged_in_user(dbsession, init.config.registry, web_server, b2, email="victim@example.com", password="secret")

    b.find_by_css("#nav-admin").click()
    b.find_by_css("#latest-user-shortcut").click()
    b.find_by_css("#btn-crud-edit").click()

    b.fill("email", "victim2@example.com")
    b.find_by_name("save").click()

    assert b.is_element_present_by_css("#msg-changes-saved")

    # Victim browser should have now logged out
    b2.visit(web_server)
    assert b2.is_element_present_by_css("#msg-session-invalidated")
    assert b2.is_element_present_by_css("#nav-sign-in")

    # See that we can log in with the new password
    b2.visit(web_server + "/login")
    b2.fill("username", "victim2@example.com")
    b2.fill("password", "secret")
    b2.find_by_name("login_email").click()

    assert b2.is_element_present_by_css("#msg-you-are-logged-in")


def test_set_enabled(browser: DriverAPI, victim_browser, web_server, init, dbsession):
    """Setting enabled resets user session. User can log after the account has been re-enabled."""
    b = browser
    b2 = victim_browser

    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)
    create_logged_in_user(dbsession, init.config.registry, web_server, b2, email="victim@example.com", password="secret")

    b.find_by_css("#nav-admin").click()
    b.find_by_css("#latest-user-shortcut").click()
    b.find_by_css("#btn-crud-edit").click()

    b.find_by_name("enabled").click()  # Turns off
    b.find_by_name("save").click()
    assert b.is_element_present_by_css("#msg-changes-saved")

    # Victim browser should have now logged out
    b2.visit(web_server)

    # We do not get session invalidated message this time, because request.user does not resolve for disabled user at all and middleware cannot distinct between anonymous and disabled user
    # assert b2.is_element_present_by_css("#msg-session-invalidated")
    assert b2.is_element_present_by_css("#nav-sign-in")

    # See that we cannot login on disabled user
    b2.visit(web_server + "/login")
    b2.fill("username", "victim@example.com")
    b2.fill("password", "secret")
    b2.find_by_name("login_email").click()
    assert b2.is_element_present_by_css("#msg-authentication-failure")

    # Re-enable the use
    b.find_by_css("#btn-crud-edit").click()
    b.find_by_name("enabled").click()  # Turns on
    b.find_by_name("save").click()
    assert b.is_element_present_by_css("#msg-changes-saved")

    # User can log in again

    # We get this message in wrong phase, but it's not really big deal as manual user deactivation should not be that common
    b2.visit(web_server)
    assert b2.is_element_present_by_css("#msg-session-invalidated")

    b2.visit(web_server + "/login")
    b2.fill("username", "victim@example.com")
    b2.fill("password", "secret")
    b2.find_by_name("login_email").click()
    assert b2.is_element_present_by_css("#msg-you-are-logged-in")


def test_delete_user_confirm(browser, web_server, init, dbsession):
    """Delete a user."""

    b = browser

    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)

    # Create another user who we are going to delete
    with transaction.manager:
        create_user(dbsession, init.config.registry, email="foobar@example.com")

    b.find_by_css("#nav-admin").click()
    b.find_by_css("#latest-user-shortcut").click()
    b.find_by_css("#btn-crud-delete").click()
    b.find_by_css("#btn-delete-yes").click()
    assert b.is_element_present_by_css("#msg-item-deleted")

    with transaction.manager:
        assert dbsession.query(User).count() == 1


def test_delete_user_cancel(browser, web_server, init, dbsession):
    """Delete a user, but back off on the confirmation screen."""

    b = browser

    create_logged_in_user(dbsession, init.config.registry, web_server, browser, admin=True)

    # Create another user who we are going to delete
    with transaction.manager:
        create_user(dbsession, init.config.registry, email="foobar@example.com")

    b.find_by_css("#nav-admin").click()
    b.find_by_css("#latest-user-shortcut").click()
    b.find_by_css("#btn-crud-delete").click()
    b.find_by_css("#btn-delete-no").click()

    # Back to the show page
    assert b.is_element_present_by_css("#crud-show")

    with transaction.manager:
        assert dbsession.query(User).count() == 2


def test_csv_export_users(dbsession, registry, browser, web_server):
    """Test CSV export functionality."""

    b = browser

    create_logged_in_user(dbsession, registry, web_server, browser, admin=True)

    unicode_bomb = "toholammin kevätkylvöt"

    with transaction.manager:
        u = dbsession.query(User).first()
        u.username = unicode_bomb

    b.find_by_css("#nav-admin").click()
    b.find_by_css("#btn-panel-list-user").click()
    assert b.is_element_present_by_css("#btn-crud-csv-export")  # This button would trigger the download of CSV that we normally cannot test with Selenium

    # Copy session cookie over to request, so we can do an authenticated user request using requests lib
    cookies = b.driver.get_cookies()

    # Convert to plain dict
    cookies = {c["name"]: c["value"] for c in cookies}
    resp = requests.get("{}/admin/models/user/csv-export".format(web_server), cookies=cookies)

    assert resp.status_code == 200
    assert resp.headers["Content-Type"] == "text/csv; charset=utf-8"
    assert unicode_bomb in resp.text



"""ws-dump-db script.

Wrapper for pgsql-dump.bash script. Extract database settings from registry and pass to Bash script.
"""
# Standard Library
import logging
import os
import subprocess
import sys
import typing as t

# Websauna
from websauna.system.devop.cmdline import init_websauna
from websauna.system.devop.exportenv import create_settings_env
from websauna.system.devop.scripts import get_config_uri
from websauna.system.devop.scripts import usage_message


DUMP_SCRIPT = os.path.join(os.path.dirname(__file__), 'psql-dump.bash')


logger = logging.getLogger(__name__)


def main(argv: t.List[str]=sys.argv):
    """Wrapper for pgsql-dump.bash script.

    :param argv: Command line arguments, second one needs to be the uri to a configuration file.
    :raises sys.SystemExit:
    """
    if len(argv) < 2:
        usage_message(
            argv,
            additional_params='[ARG1, ARG2]',
            additional_line='All arguments are passed to pg_dump command'
        )

    config_uri = get_config_uri(argv)
    request = init_websauna(config_uri)

    # Export all secrets and settings
    bash_env = create_settings_env(request.registry)

    # subprocess.check_output([DUMP_SCRIPT] + args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env)
    args = argv[2:]
    cmd = [DUMP_SCRIPT] + args

    logger.info("Running %s", " ".join(cmd))

    with subprocess.Popen(cmd, stdout=subprocess.PIPE, bufsize=1, env=bash_env, universal_newlines=True) as p:
        for line in p.stdout:
            print(line, end='')


if __name__ == "__main__":
    main()



"""Sign up form service."""
# Standard Library
import logging

# Pyramid
from pyramid.httpexceptions import HTTPFound
from pyramid.httpexceptions import HTTPNotFound
from pyramid.renderers import render_to_response
from pyramid.response import Response
from pyramid.settings import asbool
from zope.interface import implementer

# Websauna
from websauna.system.core import messages
from websauna.system.http import Request
from websauna.system.mail import send_templated_mail
from websauna.system.user.events import NewRegistrationEvent
from websauna.system.user.events import RegistrationActivatedEvent
from websauna.system.user.events import UserCreated
from websauna.system.user.interfaces import IRegistrationService
from websauna.system.user.interfaces import IUser
from websauna.system.user.utils import get_login_service
from websauna.system.user.utils import get_user_registry


logger = logging.getLogger(__name__)


@implementer(IRegistrationService)
class DefaultRegistrationService:
    """Default sign up mechanism.

    Send activation email to everybody and ask them to click a link there.
    """

    def __init__(self, request: Request):
        """Initialize registration service.

        :param request: Pyramid Request.
        """
        self.request = request

    def sign_up(self, user_data: dict) -> Response:
        """Sign up a new user.

        :param user_data: User data.
        :return: Either a redirect to a post-signup location or a page informing the user has to activate their account.
        """
        user_registry = get_user_registry(self.request)
        user = user_registry.sign_up(registration_source="email", user_data=user_data)

        # Notify site creator to initialize the admin for the first user
        self.request.registry.notify(UserCreated(self.request, user))

        settings = self.request.registry.settings

        require_activation = asbool(settings.get('websauna.require_activation', True))
        autologin = asbool(settings.get('websauna.autologin', False))

        if require_activation:
            self.create_email_activation(user)
        elif not autologin:
            messages.add(self.request, msg_id="msg-sign-up-complete", msg="Sign up complete. Welcome!", kind="success")

        self.request.registry.notify(NewRegistrationEvent(self.request, user, None, user_data))
        self.request.dbsession.flush()  # in order to get the id

        if autologin:
            login_service = get_login_service(self.request.registry)
            return login_service.authenticate(self.request, user)
        else:  # not autologin: user must log in just after registering.
            return render_to_response('login/waiting_for_activation.html', {"user": user}, request=self.request)

    def create_email_activation(self, user: IUser):
        """Create through-the-web user sign up with his/her email.

        We don't want to force the users to pick up an usernames, so we just generate an username.
        The user is free to change their username later.

        :param user: User object.
        """
        user_registry = get_user_registry(self.request)
        activation_code, expiration_seconds = user_registry.create_email_activation_token(user)

        context = {
            'link': self.request.route_url('activate', code=activation_code),
            'expiration_hours': int(expiration_seconds / 3600),
        }

        logger.info("Sending sign up email to %s", user.email)

        # TODO: Broken abstraction, we assume user.email is a attribute
        send_templated_mail(self.request, [user.email], "login/email/activate", context)

    def activate_by_email(self, activation_code: str, location: str=None) -> Response:
        """Active a user after user after the activation email.

            * User clicks link in the activation email
            * User enters the activation code on the form by hand

        :param activation_code: Activation code for user account.
        :param location: URL to redirct the user to, after activation.
        :raise: HTTPNotFound is activation_code is invalid.
        :return: Redirect to location.
        """
        request = self.request
        settings = request.registry.settings
        user_registry = get_user_registry(request)

        after_activate_url = request.route_url(settings.get('websauna.activate_redirect', 'index'))
        login_after_activation = asbool(settings.get('websauna.login_after_activation', False))

        user = user_registry.activate_user_by_email_token(activation_code)
        if not user:
            raise HTTPNotFound("Activation code not found")

        if login_after_activation:
            login_service = get_login_service(self.request.registry)
            return login_service.authenticate(self.request, user)
        else:
            self.request.registry.notify(RegistrationActivatedEvent(self.request, user, None))
            return HTTPFound(location=location or after_activate_url)



"""Default CRUD views."""
# Standard Library
import csv
import typing as t
from abc import abstractmethod
from io import StringIO

# Pyramid
import colander
import deform
from pyramid.httpexceptions import HTTPFound
from pyramid.renderers import render
from pyramid.request import Request
from pyramid.response import Response
from pyramid.view import view_config

# SQLAlchemy
from sqlalchemy.orm import Query

from slugify import slugify

# Websauna
from websauna.system.core import messages
from websauna.system.form import interstitial
from websauna.system.form.fieldmapper import EditMode
from websauna.system.form.resourceregistry import ResourceRegistry

from . import CRUD
from . import Resource
from . import paginator


class ResourceButton:
    """Present a button on the top right corner of CRUD views.

    These buttons, with breadcrumbs, form the basic navigation inside the CRUD management interface.

    Buttons are permission-aware, so they are rendered only when the user has required permission.

    The default button templates include

    * :ref:`template-crud/resource_button.html`

    * :ref:`template-crud/form_button.html`
    """

    #: The template used to render this button. Also overridable through the constructor.
    template = "crud/resource_button.html"

    def __init__(self, id: t.Optional[str]=None, name: t.Optional[str]=None, template: t.Optional[str]=None, permission: t.Optional[str]=None, tooltip: t.Optional[str]=None, feature: t.Optional[str]=None):
        """
        :param id: Id of the button to be used as HTML id
        :param name:  Human readable label of the button
        :param template: Override template for this button
        :param permission: You need to have named permission on the context to make this button visible. Set to none to make the button appear always.
        :param feature: The registry.features set must contain this feature enabled in order for the item to appear
        """

        assert id, "Button id missing"
        assert name, "Button name missing"

        self.id = id
        self.name = name

        self.permission = permission
        self.feature = feature
        self.tooltip = tooltip

        if template:
            self.template = template

    def is_visible(self, context: Resource, request: Request) -> bool:
        """Determine if we should render this button.

        :param context: Traversal context
        :param request: Current HTTP Request.
        :returns: Boolean indicating if button is visible or not.
        """
        visible = True
        if self.permission is not None:
            if not request.has_permission(self.permission, context):
                visible = False

        if self.feature is not None:
            if self.feature not in request.registry.features:
                visible = False

        return visible

    def get_link(self, context: Resource, request: Request) -> str:
        """Generate a link where this button is pointing at.

        :param context: Traversal context
        :param request: Current HTTP Request.
        :returns: Link.
        """
        return "#"

    def render(self, context: Resource, request: Request) -> str:
        """Return HTML code for this button.

        :param context: Traversal context
        :param request: Current HTTP Request.
        :returns: Rendered template.
        """
        template_context = dict(context=context, button=self)
        return render(self.template, template_context, request=request)


class TraverseLinkButton(ResourceButton):
    """A button which is a link to another page on this resource."""

    def __init__(self, view_name: str, **kwargs):
        """
        :param view_name: request.resource_url() view name where this button points at
        :param kwargs: Other ResourceButton construction arguments
        """
        super(TraverseLinkButton, self).__init__(**kwargs)
        self.view_name = view_name

    def get_link(self, context: Resource, request: Request) -> str:
        """Generate a link where this button is pointing at.

        :param context: Traversal context
        :param request: Current HTTP Request.
        :returns: Link.
        """
        return request.resource_url(context, self.view_name)


class CRUDView:
    """Base class for different CRUD views.

    This includes views having context as the listing and having context as the individual item."""

    #: base_template may point into a template providing ``crud_content`` block where the contents of the view is rendered. This allows you to decorate your CRUD which a specific page frame.
    base_template = "crud/base.html"

    #: Instance of ResourceButtons which appear on the top right corner of this view
    resource_buttons = []

    def get_resource_buttons(self) -> t.List:
        """Get the context-sensitive button options presented on this page.

        These are usually links to "Show", "Edit", "Delete" of the formm context, but you are free to add your own buttons here. The buttons are usually placed on the top right corner of the form.

        Unless you want to dynamically generate buttons depending on the page, it is enough to set ``resource_buttons`` attribute on the view class.

        :return: OrderedDict instance where key is button id and value is the link to where the button points.
        """
        return self.resource_buttons


class Listing(CRUDView):
    """List items in CRUD."""

    #: Describe what columns our listing should contain
    table = None  # type: Table

    #: How the result of this list should be split to pages
    paginator = paginator.DefaultPaginator()

    resource_buttons = [TraverseLinkButton(id="add", name="Add", view_name="add", permission="add")]

    def __init__(self, context: CRUD, request: Request):
        """Initialize Listing view.

        :param context: Points to ``CRUD`` instance.
        :param request: Current HTTP Request.
        """
        self.context = context
        self.request = request

    def get_crud(self) -> CRUD:
        return self.context

    def get_model(self) -> t.Any:
        return self.context.get_model()

    def get_query(self) -> Query:
        """Get SQLAlchemy query used in this CRUD listing.

        This can include filtering e.g. request user, crud parameters, so on.
        """
        return self.context.get_query()

    def get_count(self, query: Query):
        """Calculate total item count based on query."""
        return query.count()

    def order_query(self, query: Query):
        """Sort the query."""
        return query

    def get_title(self) -> str:
        """Get the user-readable name of the listing view (breadcrumbs, etc.)"""
        return "All {}".format(self.get_crud().plural_name)

    def paginate(self, template_context):
        """Create template variables for paginatoin results."""
        batch = self.paginator.paginate(
            template_context["query"],
            self.request,
            template_context["count"]
        )
        template_context["batch"] = batch

    @view_config(context=CRUD, name="listing", renderer="crud/listing.html", permission="view")
    def listing(self):
        """View for listing model contents in CRUD."""

        crud = self.context

        table = self.table
        columns = table.get_columns()

        # Some pre-render sanity checks

        if not columns:
            raise RuntimeError("CRUD listing doesn't not define any columns: {}".format(self.context))

        for c in columns:
            if not c.header_template:
                raise RuntimeError("header_template missing for column: {}".format(c))

        query = self.get_query()
        query = self.order_query(query)
        base_template = self.base_template

        # This is to support breadcrums with titled views
        current_view_name = self.get_title()

        title = self.get_title()
        count = self.get_count(query)

        # Base listing template variables
        template_context = {
            "title": title,
            "current_view_name": current_view_name,
            "query": query,
            "count": count,
            "crud": crud,
            "columns": columns,
            "resource_buttons": self.get_resource_buttons(),
            "base_template": base_template,
            "view": self,
        }

        # Include pagination template context
        self.paginate(template_context)

        return template_context


class CSVListing(Listing):
    """A listing view that exports the listing table as CSV.

    CSVListing users the same :py:class:`Table` structure to define the listing as the listing HTML page. For columns, we use only id and :py:meth:`websauna.system.crud.listing.Column.get_value` to stringify entries from SQLAlchemy model attributes to CSV writer stream.

    For example usage see :py:class:`websauna.system.user.adminviews.UserCSVListing`.

    .. note ::

        This is TODO, having pyramid_tm issue open https://github.com/Pylons/pyramid_tm/issues/56

    Original implementation in https://github.com/nandoflorestan/bag/blob/master/bag/spreadsheet/csv.py by Nando Florestan.
    """

    #: How many rows we buffer in a chunk before writing into a response
    buffered_rows = 100

    @view_config(context=CRUD, name="csv-export", permission='view')
    def listing(self):
        """Listing core."""

        table = self.table
        columns = table.get_columns()
        query = self.get_query()
        query = self.order_query(query)

        file_title = slugify(self.context.title)
        encoding = "utf-8"

        response = Response()
        response.headers["Content-Type"] = "text/csv; charset={}".format(encoding)
        response.headers["Content-Disposition"] = "attachment;filename={filename}.{encoding}.csv".format(
            filename=file_title,
            encoding=encoding
        )

        buf = StringIO()
        writer = csv.writer(buf)
        view = self

        def generate_csv_data():

            # Write headers
            writer.writerow([c.id for c in columns])

            # Write each listing item
            for idx, model_instance in enumerate(query):

                # Extract column values for this row
                values = [c.get_value(view, model_instance) for c in columns]

                writer.writerow(values)
                # if idx % buffered_rows == 0:
                #    yield buf.getvalue().encode(encoding)
                #    buf.truncate(0)  # But in Python 3, truncate() does not move
                #    buf.seek(0)  # the file pointer, so we seek(0) explicitly.

            # yield buf.getvalue().encode(encoding)

            # Abort the transaction, otherwise it might not be closed by underlying machinery
            # (at least tests hang)
            # TODO: Confirm this behavior with pyramid_tm 2.0 when it's out
            # request.tm.abort()

        # TODO: This use to be response.app_iter, but apparently it doesn't place nicely with pyramid_tm
        generate_csv_data()
        response.body = buf.getvalue().encode(encoding)
        return response


class FormView(CRUDView):
    """An abstract base class for form-based CRUD views.

    Use Deform form library. This views is abstract and it does not have dependency to any underlying model system like SQLAlchemy.
    """
    #: If the child class is not overriding the rendering loop, point this to a template which provides the page frame and ``crud_content`` block. For example use see :py:class:`websauna.system.user.adminviews.UserAdd`.
    base_template = "crud/base.html"

    #: This is an instance of :py:class:`websauna.system.crud.formgenerator.FormGenerator`. For SQLAlchemy models it is :py:class:`websauna.system.crud.formgenerator.SQLAlchemyFormGenerator`. Form generator describers how a CRUD model is turned to a Deform form. It is called by :py:meth:`create_form`. For example use cases see e.g. :py:class:`websauna.system.user.adminviews.UserAdd`.
    form_generator = None

    def __init__(self, context: Resource, request: Request):
        """
        :param context: Instance of ``traverse.Resource()`` or its subclasses
        :param request: HTTP request
        """
        self.context = context
        self.request = request

    def create_form(self, mode: EditMode, buttons=(), nested=None) -> deform.Form:
        model = self.get_model()
        assert getattr(self, "form_generator", None), "Class {} must define a form_generator".format(self)
        return self.form_generator.generate_form(request=self.request, context=self.context, model=model, mode=mode, buttons=buttons)

    @abstractmethod
    def get_form(self) -> deform.Form:
        """Create the form object for a view.

        Subclasses most override this, call ``create_form()`` and pass correct edit mode and buttons.
        """

    def get_crud(self) -> CRUD:
        """Get CRUD manager object for this view."""
        return self.context.__parent__

    def get_model(self) -> type:
        """Get SQLAlchemy model we are managing."""
        return self.get_crud().get_model()

    def get_object(self):
        """Get underlying SQLAlchemy model instance from current Pyramid traversing context."""
        return self.context.get_object()

    def get_title(self):
        """Get human-readable title for for template page title."""
        return "#{}".format(self.get_object().id)

    def customize_schema(self, schema: colander.Schema):
        """After Colander schema is automatically generated from the SQLAlchemy model, edit it in-place for fine-tuning.

        Override this in your view subclass for schema customizations.
        """
        pass

    def pull_in_widget_resources(self, form: deform.Form):
        """Include widget JS and CSS on the page.

        Call this as the last thing before returning template context variables from your view.
        """

        resource_registry = form.resource_registry
        assert isinstance(resource_registry, ResourceRegistry), "Websauna CRUD view got vanilla deform ResourceRegistry instance. To use widgets with dynamic JavaScript correctly you need to make sure you initialize websauna.system.form.resourceregistry.ResourceRegistry for your form object."

        resource_registry.pull_in_resources(self.request, form)


class Show(FormView):
    """Show one instance of a model."""

    resource_buttons = [TraverseLinkButton(id="edit", name="Edit", view_name="edit", permission="edit")]

    def get_title(self):
        return self.context.get_title()

    def get_form(self):
        return self.create_form(EditMode.show, buttons=())

    def get_form_context(self) -> object:
        """Get the item that populates the form."""
        return self.context.get_object()

    def get_appstruct(self, form, form_context) -> dict:
        """Get the dictionary that populates the form."""
        return form.schema.dictify(form_context)

    @view_config(context=Resource, name="show", renderer="crud/show.html", permission='view')
    def show(self):
        """View for showing an individual object."""
        obj = self.get_object()
        form_context = self.get_form_context()
        base_template = self.base_template

        form = self.get_form()
        appstruct = self.get_appstruct(form, form_context)
        rendered_form = form.render(appstruct, readonly=True)

        resource_buttons = self.get_resource_buttons()

        crud = self.get_crud()

        title = self.get_title()
        self.pull_in_widget_resources(form)
        return dict(form=rendered_form, context=self.context, obj=obj, title=title, crud=crud, base_template=base_template, resource_buttons=resource_buttons)


class Edit(FormView):
    """Edit model instance using Deform form..

    The call order of functions

    * ``edit()``

    * ``save_changes()``

    * ``do_success()``
    """

    resource_buttons = [TraverseLinkButton(id="show", name="Show", view_name="show", permission="view")]

    def get_title(self):
        return "Editing #{}".format(self.context.get_title())

    def get_buttons(self) -> t.Iterable[deform.form.Button]:
        return (
            deform.form.Button("save"),
            deform.form.Button("cancel"),
        )

    def get_form(self) -> deform.form.Form:
        """Get a form used to edit this item."""
        return self.create_form(EditMode.edit, buttons=self.get_buttons())

    def bind_schema(self, schema):
        return schema.bind(obj=self.context.get_object(), request=self.request, context=self.context)

    def do_success(self) -> Response:
        """Called after the save (objectify) has succeeded."""
        messages.add(self.request, kind="success", msg="Changes saved.", msg_id="msg-changes-saved")

        # Redirect back to view page after edit page has succeeded
        return HTTPFound(self.request.resource_url(self.context, "show"))

    def do_cancel(self) -> Response:
        """Called when user presses the cancel button.

        :return: HTTPResponse
        """

        # Redirect back to view page after edit page has succeeded
        return HTTPFound(self.request.resource_url(self.context, "show"))

    def save_changes(self, form: deform.Form, appstruct: dict, obj: object):
        """Store the data from the form on the object."""
        form.schema.objectify(appstruct, obj)

    def get_appstruct(self, form: deform.Form, obj: object) -> dict:
        """Turn the object to form editable format."""
        return form.schema.dictify(obj)

    @view_config(context=Resource, name="edit", renderer="crud/edit.html", permission='edit')
    def edit(self):
        """View for showing an individual object."""

        # SQLAlchemy model instance
        obj = self.context.get_object()
        base_template = self.base_template

        # Create form, convert instance to Colander structure for Deform
        form = self.get_form()

        crud = self.get_crud()

        title = self.get_title()

        if "save" in self.request.POST:

            controls = self.request.POST.items()

            try:
                appstruct = form.validate(controls)
                self.save_changes(form, appstruct, obj)
                return self.do_success()

            except deform.ValidationFailure as e:
                # Whoops, bad things happened, render form with validation errors
                rendered_form = e.render()

        elif "cancel" in self.request.POST:
            # User pressed cancel
            return self.do_cancel()
        else:
            # Render initial form view with populated values
            appstruct = self.get_appstruct(form, obj)
            rendered_form = form.render(appstruct)

        self.pull_in_widget_resources(form)

        return dict(form=rendered_form, context=self.context, obj=obj, title=title, crud=crud, base_template=base_template, resource_buttons=self.get_resource_buttons(), current_view_name="Edit")


class Add(FormView):
    """Create a new item in CRUD."""

    def get_title(self):
        return "Add new {}".format(self.get_crud().singular_name)

    def get_form(self) -> object:
        return self.create_form(EditMode.add, buttons=self.get_buttons())

    def get_crud(self) -> CRUD:
        """Get CRUD manager object for this view."""
        return self.context

    def get_model(self) -> type:
        return self.get_crud().get_model()

    def create_object(self):
        """Factory method to create a new empty object to be populated from the form."""
        model = self.get_model()
        return model()

    def initialize_object(self, form, appstruct: dict, obj: object):
        """Record values from the form on a freshly created object."""
        form.schema.objectify(appstruct, obj)

    def bind_schema(self, schema):
        return schema.bind(request=self.request, context=self.context)

    def add_object(self, obj):
        """Add objects to transaction lifecycle and flush newly created object to persist storage to give them id."""
        dbsession = self.context.get_dbsession()
        dbsession.add(obj)
        dbsession.flush()

    def do_success(self, resource: Resource) -> Response:
        """Finish action after saving new object.

        Usually returns HTTP redirect to the next view.
        """
        messages.add(self.request, kind="success", msg="Item added.", msg_id="msg-item-added")
        # Redirect back to view page after edit page has succeeded
        return HTTPFound(self.request.resource_url(resource, "show"))

    def do_cancel(self) -> Response:
        """Called when user presses the cancel button.

        :return: HTTPResponse
        """
        return HTTPFound(self.request.resource_url(self.get_crud(), "listing"))

    def build_object(self, form, appstruct: dict) -> object:
        """Builds a new object.

        The default behavior is to call :py:meth:`create_object` to construct a new object, then populate it with :py:meth:`initialize_object` and finally include the created object in the transaction lifecycle with :py:meth:`add_object`.
        """
        obj = self.create_object()
        self.initialize_object(form, appstruct, obj)
        # We do not need to explicitly call save() or commit() as we are using Zope transaction manager
        self.add_object(obj)
        return obj

    def get_buttons(self)-> t.List[deform.form.Button]:
        buttons = (
            deform.form.Button("add"),
            deform.form.Button("cancel"),
        )
        return buttons

    @view_config(context=CRUD, name="add", renderer="crud/add.html", permission='add')
    def add(self):
        """View for showing an individual object."""

        # SQLAlchemy model instance
        base_template = self.base_template

        # Create form, convert instance to Colander structure for Deform
        form = self.get_form()

        crud = self.get_crud()

        title = self.get_title()

        if "add" in self.request.POST:

            controls = self.request.POST.items()

            try:
                appstruct = form.validate(controls)

                # Cannot update id, as it is read-only
                if 'id' in appstruct:
                    del appstruct["id"]

                obj = self.build_object(form, appstruct)
                resource = crud.wrap_to_resource(obj)

                return self.do_success(resource)

            except deform.ValidationFailure as e:
                # Whoops, bad things happened, render form with validation errors
                rendered_form = e.render()

        elif "cancel" in self.request.POST:
            # User pressed cancel
            return self.do_cancel()
        else:
            # Render initial form view with populated values
            rendered_form = form.render()

        self.pull_in_widget_resources(form)

        return dict(form=rendered_form, context=self.context, title=title, crud=crud, base_template=base_template, resource_buttons=self.get_resource_buttons(), current_view_name="Add")


class Delete:
    """Delete one item within a CRUD with a confirmation screen.

    This is an abstract item delete implementation; you must either set :py:attr:`deleter` callback or override :py:meth:`yes`.
    """
    base_template = "crud/base.html"

    #: callback ``deleter(request, context)`` which is called to perform the actual model specific delete operation.
    deleter = None

    def __init__(self, context: Resource, request: Request):
        self.context = context
        self.request = request

    def get_crud(self) -> CRUD:
        return self.context.__parent__

    def get_model(self) -> object:
        return self.context.get_model()

    def get_object(self):
        """Get underlying SQLAlchemy model instance from current Pyramid traversing context."""
        return self.context.get_object()

    def delete_object(self):
        """User picked YEAH LET'S DO IT.

        First try to call ``delete`` callback if one is set. If not then fallback to :py:meth:`websauna.system.crud.CRUD.delete_object`.

        http://opensourcehacker.com/wp-content/uploads/2013/04/koala.gif
        """

        if self.deleter:
            self.deleter(self.context, self.request)
        else:
            self.get_crud().delete_object(self.get_object())

        messages.add(self.request, "Deleted {}".format(self.context.get_title()), msg_id="msg-item-deleted", kind="success")

        return HTTPFound(self.request.resource_url(self.get_crud(), "listing"))

    def cancel_delete(self):
        """Permanent data destruction is bad.

        Redirect user back to the show view.
        """
        return HTTPFound(self.request.resource_url(self.context, "show"))

    @view_config(context=Resource, name="delete", renderer="crud/delete.html", permission='delete')
    def delete(self):
        """Delete view endpoint."""
        choices = (
            interstitial.Choice("Yes", self.delete_object, id="btn-delete-yes"),
            interstitial.Choice("No", self.cancel_delete, id="btn-delete-no"),
        )

        if self.request.method == "POST":
            return interstitial.process_interstitial(self.request, choices)

        # Expose base_template to the template rendering
        base_template = self.base_template

        current_view_name = "Confirm delete"

        return locals()



"""Test the creation of a new websauna application, using cookiecutter templates."""
# Standard Library
import os

import pytest
from flaky import flaky

from .scaffold import app_scaffold  # noQA
from .scaffold import cookiecutter_config  # noQA
from .scaffold import create_psq_db
from .scaffold import execute_venv_command
from .scaffold import replace_file
from .scaffold import start_ws_pserve


pytestmark = pytest.mark.skip(reason="Tests failing due to problem in automated creation of  virtualenv")
# see https://github.com/pypa/virtualenv/issues/228


@pytest.fixture()
def dev_db(request, ini_settings):
    """Create PostgreSQL database myapcdp_dev and return its connection information."""
    dsn = ini_settings["sqlalchemy.url"]
    if dsn.startswith("sqlite://"):
        pytest.skip("These tests are run only for PostgreSQL database")
    return create_psq_db(request, "app_dev", dsn)


def test_pserve(app_scaffold, dev_db, browser):  # noQA
    """Create an application and see if ws-pserve starts. """
    # User models are needed to start the web server
    execute_venv_command("cd my.app && ws-sync-db ws://my/app/conf/development.ini", app_scaffold)
    server = start_ws_pserve("cd my.app && ws-pserve ws://my/app/conf/development.ini", app_scaffold)
    try:
        # Make sure we get some sensible output from the server
        b = browser
        b.visit("http://localhost:6543")
        # See our scaffold home page loads and demo text is there
        assert b.is_element_present_by_css("#demo-text")
    finally:
        server.kill()


# On Travis there might be abnormal delays in this test
@flaky  # noQA
def test_pyramid_debugtoolbar(app_scaffold, dev_db, browser):
    """Pyramid debug toolbar should be effective with the default development ws-pserve."""
    # User models are needed to start the web server
    execute_venv_command("ws-sync-db ws://my/app/conf/development.ini", app_scaffold, cd_folder="my.app")
    server = start_ws_pserve("cd my.app && ws-pserve ws://my/app/conf/development.ini", app_scaffold)
    try:
        # Make sure we get some sensible output from the server
        b = browser
        b.visit("http://localhost:6543/error-trigger")
        # See that Wertzkreug debugger is up
        assert b.is_element_present_by_css(".traceback")
    finally:
        server.kill()


def test_sdist(app_scaffold):  # noQA
    """Create an application and see if sdist egg can be created. """
    execute_venv_command("cd my.app && python setup.py sdist", app_scaffold)


def test_app_syncdb(app_scaffold, dev_db):  # noQA
    """Create an application and see if database is correctly created."""
    execute_venv_command("cd my.app && ws-sync-db ws://my/app/conf/development.ini", app_scaffold)


def test_shell(app_scaffold, dev_db):  # noQA
    """Create an application and see if shell is started."""
    execute_venv_command("cd my.app && ws-shell ws://my/app/conf/development.ini", app_scaffold, wait_and_see=2.0)


def test_db_shell(app_scaffold, dev_db):  # noQA
    """Create an application and see if shell is started."""
    execute_venv_command("cd my.app && ws-db-shell ws://my/app/conf/development.ini", app_scaffold, wait_and_see=2.0)


def test_gitignore(app_scaffold, dev_db):  # noQA
    """We construct a proper .gitignore file for the project."""
    assert os.path.exists(os.path.join(app_scaffold, "my.app", ".gitignore"))


def test_migration(app_scaffold, dev_db):  # noQA
    """Create an application, add a model and see if migrations are run."""
    with replace_file(os.path.join(app_scaffold, "my.app", "my", "app", "models.py"), MODELS_PY):
        # Happens if py.test picks up this test twice
        files = os.listdir(os.path.join(app_scaffold, "my.app", "alembic", "versions"))
        for f in files:
            assert not f.endswith(".py"), "versions folder contained migration scripts before initial Alembic run: {}".format(files)
        execute_venv_command("cd my.app && ws-alembic -x packages=all -c my/app/conf/development.ini revision --autogenerate -m 'Added MyModel'", app_scaffold)
        execute_venv_command("cd my.app && ws-alembic -x packages=all -c my/app/conf/development.ini upgrade head", app_scaffold)
        # Assert we got migration script for mymodel
        files = os.listdir(os.path.join(app_scaffold, "my.app", "alembic", "versions"))
        assert any("added_mymodel.py" in f for f in files), "Got files {}".format(files)


def test_app_sanity_check_fail(app_scaffold, dev_db):  # noQA
    """Create an application and see we don't start if migrations are not run."""
    execute_venv_command("cd my.app && ws-pserve ws://my/app/conf/development.ini", app_scaffold, assert_exit=1)


# DB dump on Travis might hung?
@flaky  # noQA
def test_dump_db(app_scaffold, dev_db):
    """Test database dump."""
    execute_venv_command("cd my.app && ws-sync-db ws://my/app/conf/development.ini", app_scaffold)
    execute_venv_command("cd my.app && ws-dump-db ws://my/app/conf/development.ini", app_scaffold)


# Browser doesn't come up timely on Travis
@flaky  # noQA
def test_create_user(app_scaffold, dev_db, browser):
    """Test creating user from command line and logging in as this user."""
    execute_venv_command("ws-sync-db ws://my/app/conf/development.ini", app_scaffold, cd_folder="my.app")
    exitcode, stdout, stderr = execute_venv_command("ws-create-user ws://my/app/conf/development.ini mikko@example.com secret", app_scaffold, cd_folder="my.app")
    print("Got ws-create-usr ", stdout, stderr)
    server = start_ws_pserve("cd my.app && ws-pserve ws://my/app/conf/development.ini", app_scaffold)
    try:
        # Make sure we get some sensible output from the server
        b = browser
        b.visit("http://localhost:6543/login")
        assert b.is_element_present_by_css("#login-form")
        # See our scaffold home page loads and demo text is there
        b.fill("username", "mikko@example.com")
        b.fill("password", "secret")
        b.find_by_name("login_email").click()
        assert b.is_element_present_by_css("#nav-admin")
    finally:
        server.kill()


def test_tweens(app_scaffold, dev_db):  # noQA
    """Test tweens command."""
    # TODO: Tweens should not really depent on database, but let's fix this later
    execute_venv_command("cd my.app && ws-sync-db ws://my/app/conf/development.ini", app_scaffold)
    execute_venv_command("cd my.app && ws-tweens ws://my/app/conf/development.ini", app_scaffold)


def test_create_table(app_scaffold, dev_db):  # noQA
    """Test ws-create-table command."""
    execute_venv_command("ws-create-table ws://my/app/conf/development.ini", app_scaffold, cd_folder="my.app")


def test_sanity_check(app_scaffold, dev_db):  # noQA
    """Test sanity check command."""
    execute_venv_command("ws-sanity-check ws://my/app/conf/development.ini", app_scaffold, cd_folder="my.app", assert_exit=10)
    execute_venv_command("ws-sync-db ws://my/app/conf/development.ini", app_scaffold, cd_folder="my.app")
    execute_venv_command("ws-sanity-check ws://my/app/conf/development.ini", app_scaffold, cd_folder="my.app", assert_exit=0)


#: Migration test file
MODELS_PY = """
from websauna.system.model.meta import Base
from sqlalchemy import Column
from sqlalchemy import Integer
class MyModel(Base):
    __tablename__ = "mymodel"
    id = Column(Integer, autoincrement=True, primary_key=True)
"""



"""User utilities."""
# Standard Library
import typing as t

# Pyramid
from pyramid.interfaces import IRequest
from pyramid.registry import Registry

from authomatic import Authomatic

# Websauna
from websauna.system.http import Request
from websauna.system.user.interfaces import IActivationModel
from websauna.system.user.interfaces import IAuthomatic
from websauna.system.user.interfaces import ICredentialActivityService
from websauna.system.user.interfaces import IGroupModel
from websauna.system.user.interfaces import ILoginService
from websauna.system.user.interfaces import IOAuthLoginService
from websauna.system.user.interfaces import IRegistrationService
from websauna.system.user.interfaces import ISiteCreator
from websauna.system.user.interfaces import ISocialLoginMapper
from websauna.system.user.interfaces import IUserModel
from websauna.system.user.interfaces import IUserRegistry


def get_user_class(registry: Registry) -> t.Type[IUserModel]:
    """Get the class implementing IUserModel.

    :param registry: Pyramid registry.
    :return: Class implementing IUserModel.
    """
    user_class = registry.queryUtility(IUserModel)
    return user_class


def get_group_class(registry: Registry) -> t.Type[IGroupModel]:
    """Get the class implementing IGroupModel.

    :param registry: Pyramid registry.
    :return: Class implementing IGroupModel.
    """
    group_class = registry.queryUtility(IGroupModel)
    return group_class


def get_activation_model(registry: Registry) -> t.Type[IActivationModel]:
    """Get the class implementing IActivationModel.

    :param registry: Pyramid registry.
    :return: Class implementing IActivationModel.
    """
    activation_model = registry.queryUtility(IActivationModel)
    return activation_model


def get_site_creator(registry: Registry) -> ISiteCreator:
    """Get the class implementing ISiteCreator.

    :param registry: Pyramid registry.
    :return: Class implementing ISiteCreator.
    """
    site_creator = registry.queryUtility(ISiteCreator)
    return site_creator


def get_authomatic(registry: Registry) -> Authomatic:
    """Get active Authomatic instance from the registry.

    This is registered in ``Initializer.configure_authomatic()``.
    :param registry: Pyramid registry.
    :return: Instance of Authomatic.
    """
    instance = registry.queryUtility(IAuthomatic)
    return instance


def get_social_login_mapper(registry: Registry, provider_id: str) -> ISocialLoginMapper:
    """Get a named social login mapper.

    Example::

        get_social_login_mapper(registry, "facebook")

    :param registry: Pyramid registry.
    :param provider_id: Provider id of a social login mapper.
    :return: Implementation of ISocialLoginMapper.
    """
    return registry.queryUtility(ISocialLoginMapper, name=provider_id)


def get_login_service(request: Request) -> ILoginService:
    """Get the login service.

    :param request: Pyramid request.
    :return: Implementation of ILoginService.
    """
    assert IRequest.providedBy(request)
    return request.registry.queryAdapter(request, ILoginService)


def get_oauth_login_service(request: Request) -> IOAuthLoginService:
    """Get the oauth login service.

    :param request: Pyramid request.
    :return: Implementation of IOAuthLoginService.
    """
    assert IRequest.providedBy(request)
    return request.registry.queryAdapter(request, IOAuthLoginService)


def get_user_registry(request: Request) -> IUserRegistry:
    """Get the user registry..

    :param request: Pyramid request.
    :return: Implementation of IUserRegistry.
    """
    return request.registry.queryAdapter(request, IUserRegistry)


def get_credential_activity_service(request: Request) -> ICredentialActivityService:
    """Get the credential activity service.

    :param request: Pyramid request.
    :return: Implementation of ICredentialActivityService.
    """
    assert IRequest.providedBy(request)
    return request.registry.queryAdapter(request, ICredentialActivityService)


def get_registration_service(request: Request) -> IRegistrationService:
    """Get the registration service.

    :param request: Pyramid request.
    :return: Implementation of IRegistrationService.
    """
    assert IRequest.providedBy(request)
    return request.registry.queryAdapter(request, IRegistrationService)



"""Tutorial models and admins for CRUD testing."""

# Standard Library
import datetime
from uuid import uuid4

# SQLAlchemy
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.orm import relationship

# Websauna
from websauna.system.admin.modeladmin import ModelAdmin
from websauna.system.admin.modeladmin import model_admin
from websauna.system.model.columns import UUID
from websauna.system.model.columns import UTCDateTime
from websauna.system.model.meta import Base
from websauna.utils.time import now


class Question(Base):

    #: The table in the database
    __tablename__ = "question"

    #: Database primary key for the row (running counter)
    id = Column(Integer, autoincrement=True, primary_key=True)

    #: Publicly exposed non-guessable
    uuid = Column(UUID(as_uuid=True), default=uuid4)

    #: Question text
    question_text = Column(String(256), default=None)

    #: When this question was published
    published_at = Column(UTCDateTime, default=None)

    #: Relationship mapping between question and choice.
    #: Each choice can have only question.
    #: Deleteing question deletes its choices.
    choices = relationship("Choice",
                           back_populates="question",
                           lazy="dynamic",
                           cascade="all, delete-orphan",
                           single_parent=True)

    def is_recent(self):
        return self.published_at >= now() - datetime.timedelta(days=1)

    def __repr__(self):
        return "#{}: {}".format(self.id, self.question_text)

    def __str__(self):
        """Python default and admin UI string presentation."""
        return self.question_text


class Choice(Base):

    #: The table in the database
    __tablename__ = "choice"

    #: Database primary key for the row (running counter)
    id = Column(Integer, autoincrement=True, primary_key=True)

    #: Publicly exposed non-guessable id
    uuid = Column(UUID(as_uuid=True), default=uuid4)

    #: What the user sees for this choice
    choice_text = Column(String(256), default=None)

    #: How many times this choice has been voted
    votes = Column(Integer, default=0)

    #: Which question this choice is part of
    question_id = Column(Integer, ForeignKey('question.id'))
    question = relationship("Question", back_populates="choices", uselist=False)

    def __repr__(self):
        """Shell and debugger presentation."""
        return "#{}: {}".format(self.id, self.choice_text)

    def __str__(self):
        """Python default and admin UI string presentation."""
        return self.choice_text


@model_admin(traverse_id="question")
class QuestionAdmin(ModelAdmin):

    title = "Questions"

    singular_name = "question"
    plural_name = "questions"
    model = Question

    class Resource(ModelAdmin.Resource):

        def get_title(self):
            return self.get_object().question_text


@model_admin(traverse_id="choice")
class ChoiceAdmin(ModelAdmin):

    title = "Choices"

    singular_name = "choice"
    plural_name = "choices"
    model = Choice

    class Resource(ModelAdmin.Resource):

        def get_title(self):
            return self.get_object().choice_text



"""Test tasks for the scheduler."""
# Standard Library
import logging

# Websauna
from websauna.system import DemoInitializer
from websauna.system.core.redis import get_redis
from websauna.system.task.tasks import RetryableTransactionTask
from websauna.system.task.tasks import ScheduleOnCommitTask
from websauna.system.task.tasks import WebsaunaTask
from websauna.system.task.tasks import task
from websauna.system.user.models import User


logger = logging.getLogger(__name__)


# Configured to be executed every second in scheduler-test.ini
@task(name="foobar", bind=True, base=WebsaunaTask)
def redis_test_write(self: WebsaunaTask):
    logger.error("Called by beat")
    request = self.get_request()
    connection = get_redis(request)
    connection.set("foo", "xoo")


@task(name="crashaxs")
def crash():
    raise RuntimeError("Test exception")


@task(base=RetryableTransactionTask, bind=True)
def modify_username(self, user_id):
    """A demo task we use to detect that Celery tasks write database changes correctly."""
    request = self.request.request
    dbsession = request.dbsession

    u = dbsession.query(User).get(user_id)
    u.username = "set by celery"


@task(base=RetryableTransactionTask)
def modify_username_abort_by_exception(self, user_id):
    """We never commit as exception is raisen."""
    request = self.request.request
    dbsession = request.dbsession

    u = dbsession.query(User).get(user_id)
    u.username = "set by celery"  # This change should not commit
    raise RuntimeError("Exception raised")


@task(base=ScheduleOnCommitTask, bind=True)
def modify_username_manual_transaction(self, user_id):
    """Example of a manual transaction management within a task."""
    request = self.request.request

    with request.tm:

        dbsession = request.dbsession

        u = dbsession.query(User).get(user_id)
        u.username = "set by celery"


class SchedulerInitializer(DemoInitializer):
    """Entry point for tests stressting task functionality."""

    def configure_tasks(self):
        self.config.scan("websauna.tests.demotasks")


def main(global_config, **settings):
    return SchedulerInitializer.bootstrap(global_config)



"""An abstract CRUD implementation based on traversal. The default support for SQLAlchemy and Deform."""
# Standard Library
import typing as t
from abc import abstractmethod

# Pyramid
from pyramid.interfaces import IRequest

# Websauna
from websauna.system.core.traversal import Resource as _Resource
from websauna.system.http import Request

from .urlmapper import Base64UUIDMapper


class Resource(_Resource):
    """One object in CRUD traversing.

    Maps the raw database object under CRUD view/edit/delete control to traverse path.

    Presents an underlying model instance mapped to an URL path. ``__parent__`` attribute points to a CRUD instance.
    """

    def __init__(self, request: Request, obj: object):
        """Initialize resource object.

        :param obj: The underlying object we wish to wrap for traversing. Usually SQLAlchemy model instance.
        """
        # Some safety checks we get arguments correctly.n
        assert IRequest.providedBy(request)

        self.request = request
        self.obj = obj

    def get_object(self) -> t.Any:
        """Return the wrapped database object."""
        return self.obj

    def get_path(self):
        """Extract the traverse path name from the database object."""
        assert hasattr(self, "__parent__"), "get_path() can be called only for objects whose lineage is set by make_lineage()"

        crud = self.__parent__
        path = crud.mapper.get_path_from_object(self.obj)
        return path

    def get_model(self) -> type:
        """Get the model class represented by this resource."""
        return self.__parent__.get_model()

    def get_title(self) -> str:
        """Title used on view, edit, delete, pages.

        By default use the capitalized URL path path.
        """
        return self.get_path()


class CRUD(_Resource):
    """Define create-read-update-delete interface for an model.

    We use Pyramid traversing to get automatic ACL permission support for operations. As long given CRUD resource parts define __acl__ attribute, permissions are respected automatically.

    URLs are the following:

        List: $base/listing

        Add: $base/add

        Show $base/$id/show

        Edit: $base/$id/edit

        Delete: $base/$id/delete
    """
    # How the model is referred in templates. e.g. "User"
    title = "(untitled CRUD)"

    # TODO: This is inperfect directly copied from Django - many languages have more plural forms than two. Fix when i18n lands
    #: Helper noun used in the default placeholder texts
    singular_name = "item"

    # TODO: This is inperfect directly copied from Django - many languages have more plural forms than two. Fix when i18n lands
    #: Helper noun used in the default placeholder texts
    plural_name = "items"

    #: Mapper defines how objects are mapped to URL space. The default mapper assumes models have attribute ``uuid`` which is base64 encoded to URL. You can change this to :py:class:`websauna.system.crud.urlmapper.IdMapper` if you instead to want to use ``id`` as a running counter primary column in URLs. This is not recommended in security wise, though.
    mapper = Base64UUIDMapper()

    def make_resource(self, obj: object) -> Resource:
        """Take raw model instance and wrap it to Resource for traversing.

        :param obj: SQLAlchemy object or similar model object.
        :return: :py:class:`websauna.core.traverse.Resource`
        """

        # Use internal Resource class to wrap the object
        if hasattr(self, "Resource"):
            return self.Resource(self.request, obj)

        raise NotImplementedError("Does not know how to wrap to resource: {obj}".format(obj=obj))

    def wrap_to_resource(self, obj: object) -> Resource:
        """Wrap object to a traversable part.

        :param obj: SQLAlchemy object or similar model object.
        :return: :py:class:`websauna.core.traverse.Resource`
        """
        instance = self.make_resource(obj)

        path = self.mapper.get_path_from_object(obj)
        assert type(path) == str, "Object {obj} did not map to URL path correctly, got path {path}".format(obj=obj, path=path)
        instance.make_lineage(self, instance, path)
        return instance

    def traverse_to_object(self, path: str) -> Resource:
        """Wraps object to a traversable URL.

        Loads raw database object with id and puts it inside ``Instance`` object, with ``__parent__`` and ``__name__`` pointers.

        :param path: Path to be traversed to.
        :return: :py:class:`websauna.core.traverse.Resource`
        """
        # First try if we get an view for the current instance with the name
        id = self.mapper.get_id_from_path(path)
        obj = self.fetch_object(id)
        return self.wrap_to_resource(obj)

    @abstractmethod
    def fetch_object(self, id: str) -> object:
        """Load object from the database for CRUD path for view/edit/delete.

        :param id: Object id.
        :return: Object from database.
        """
        raise NotImplementedError('Please use concrete subclass like websauna.syste.crud.sqlalchemy')

    def get_object_url(self, obj: object, view_name: t.Optional[str]=None) -> str:
        """Get URL for view for an object inside this CRUD.

        :param obj: Raw object, e.g. SQLAlchemy instance, which can be wrapped with ``wrap_to_resource``.
        :param view_name: Traverse view name for the resource. E.g. ``show``, ``edit``.
        :return: URL to the object.
        """
        res = self.wrap_to_resource(obj)
        args = [res, ]
        if view_name:
            args.append(view_name)
        return self.request.resource_url(*args)

    def delete_object(self, obj: object):
        """Delete one item in the CRUD.

        Called by delete view if no alternative logic is implemented.
        :param obj: Raw object, e.g. SQLAlchemy instance.
        """
        raise NotImplementedError("The subclass must implement actual delete method or give deleter callback in Delete view.")

    def __getitem__(self, path: str) -> Resource:
        """Traverse to a model instance.

        :param path: Part of URL which is resolved to an object via ``mapper``.
        :return: :py:class:`websauna.core.traverse.Resource`
        """
        if not self.mapper.is_id(path):
            # Signal that this id is not part of the CRUD database and may be a view
            raise KeyError

        return self.traverse_to_object(path)



"""Base user forms."""
# Pyramid
from deform import Button
from deform import Form


class DefaultUserForm(Form):
    """Form implementation used for all user forms.

    Comes with one submit button.
    """

    def __init__(self, *args, **kwargs):
        """Initialize DefaultUserForm."""
        if not kwargs.get('buttons'):
            kwargs['buttons'] = ('submit',)
        super().__init__(*args, **kwargs)


class DefaultLoginForm(Form):
    """Form used by default Login by email screen."""

    def __init__(self, *args, **kwargs):
        """Initialize DefaultLoginForm."""
        login_button = Button(name="login_email", title="Login with email", css_class="btn-lg btn-block")
        kwargs['buttons'] = (login_button,)
        super().__init__(*args, **kwargs)


class DefaultRegisterForm(Form):
    """Form used by default Sign up by email screen."""

    def __init__(self, *args, **kwargs):
        """Initialize DefaultRegisterForm."""
        sign_up_button = Button(name="sign_up", title="Sign up with email", css_class="btn-lg btn-block")
        kwargs['buttons'] = (sign_up_button,)
        super().__init__(*args, **kwargs)



# Pyramid
import transaction

# Websauna
from websauna.system.user.models import Group
from websauna.system.user.models import User
from websauna.system.user.utils import get_site_creator


def test_virgin_init_admin(init, dbsession):
    """When the user logs in first time see we create admin group and permissions for it."""

    with transaction.manager:
        u = User(username="example", email="example@example.com")
        dbsession.add(u)
        dbsession.flush()
        assert not u.is_admin()

        site_creator = get_site_creator(init.config.registry)
        site_creator.init_empty_site(dbsession, u)

    with transaction.manager:
        u = dbsession.query(User).get(1)
        assert dbsession.query(Group).count() == 1
        assert u.is_admin()



"""ws-db-shell script.

Run pgcli shell on the configured database.
"""
# Standard Library
import os
import sys
import typing as t
from shutil import which

# Websauna
from websauna.system.devop.cmdline import init_websauna
from websauna.system.devop.scripts import feedback
from websauna.system.devop.scripts import feedback_and_exit
from websauna.system.devop.scripts import get_config_uri
from websauna.system.devop.scripts import usage_message


def main(argv: t.List[str]=sys.argv):
    """Run pgcli shell on the database specified on the configuration file.

    :param argv: Command line arguments, second one needs to be the uri to a configuration file.
    """
    if len(argv) < 2:
        usage_message(argv, additional_params='[var=value]')

    config_uri = get_config_uri(argv)
    request = init_websauna(config_uri)
    url = request.registry.settings.get('sqlalchemy.url')

    engine = request.dbsession.get_bind()

    if not which('pgcli'):
        message = 'pgcli is not installed\nPlease install Websauna as pip install websauna[utils] to get this dependency'
        feedback_and_exit(message, display_border=False)

    feedback('Connecting to {engine}'.format(engine=engine), display_border=False)
    os.system('pgcli {url}'.format(url=url))



"""Get Celery instance from Pyramid configuration."""
# Standard Library
import textwrap

# Pyramid
from pyramid.registry import Registry

# Celery
from celery import Celery


def parse_celery_config(celery_config_python: str) -> dict:
    # Expose timedelta object for config to be used in beat schedule
    # http://docs.celeryproject.org/en/master/userguide/periodic-tasks.html#beat-entries
    from datetime import timedelta  # noqa
    from celery.schedules import crontab  # noqa

    _globals = globals().copy()
    _locals = locals().copy()
    code = textwrap.dedent(celery_config_python)

    try:
        config_dict = eval(code, _globals, _locals)
    except Exception as e:
        raise RuntimeError("Could not execute Python code to produce Celery configuration object: {}".format(code)) from e

    if "broker_url" not in config_dict:
        raise RuntimeError("Mandatory broker_url Celery setting missing. Did we fail to parse config? {}".format(config_dict))

    return config_dict


def get_celery_config(registry: Registry) -> dict:
    """Load Celery configuration from settings.

    You need to have a setting key ``celery_config_python``. This is Python code to configure Celery. The code is executed and all locals are passed to Celery app.

    More information:

        * http://docs.celeryproject.org/en/master/userguide/configuration.html

    :param registry: Pyramid registry from where we read the Celery configuratino
    :return: An object holding Celery configuration variables
    """

    celery_config_python = registry.settings.get('websauna.celery_config')
    if not celery_config_python:
        raise RuntimeError('Using Celery with Websauna requires you to have celery_config_python configuration variable')

    return parse_celery_config(celery_config_python)


def get_celery(registry: Registry):
    """Load and configure Celery app.

    Cache the loaded Celery app object on registry.

    :param registry: Use registry settings to load Celery
    """
    celery = getattr(registry, "celery", None)
    if not celery:
        celery = registry.celery = Celery()
        celery.conf.update(get_celery_config(registry))

        # Expose Pyramid registry to Celery app and tasks
        celery.registry = registry

    return celery



"""Registering static assets with cache busting policies.

For more information read :ref:`static` narrative.
"""
# Standard Library
import hashlib
import json
import logging
import os
import shutil
import typing as t
from abc import ABC
from abc import abstractmethod
from collections import defaultdict
from os import scandir

# Pyramid
from pyramid.config import Configurator
from pyramid.decorator import reify
from pyramid.path import AssetResolver
from pyramid.request import Request

# Websauna
from websauna.system.compat.typing import DirEntry


logger = logging.getLogger(__name__)


MARKER_FOLDER = "perma-asset"

#: Generated at the root of each perma-asset folder
MANIFEST_FILE = ".manifest.json"


class StaticAssetPolicy(ABC):
    """A helper class to add static views and apply a configured cache busting policy on them."""

    def __init__(self, config: Configurator):
        """Initialize the StaticAssetPolicy.

        :param config: Pyramid configuration.
        """
        self.config = config

    @abstractmethod
    def add_static_view(self, name: str, path: str):
        """Include a path in static assets and configures cache busting for it.

        This does not only include the static resources in the routing, but sets the default cache busting policy for them in the :term:`production` environment.

        See :py:meth:`pyramid.config.Configurator.add_static_view` for more details.

        :param name: Name of the asset.
        :param path: Path to the asset.
        """


# Courtesy of http://stackoverflow.com/a/3431838/315168
def md5(filename: str) -> str:
    """Generate the md5 hash for a file with given filename.

    :param filename: Name of the file to generate the MD5 hash for.
    :return: md5 hash of the file.
    """
    hash = hashlib.md5()
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash.update(chunk)
    return hash.hexdigest()


class CopyAndHashCollector:
    """Toss all static files into perma-asset folder, MD5 hash included in the name."""

    def __init__(self, root: str, settings: dict, target_path: str=None):
        """Initialize CopyAndHashCollector.

        :param root: Root path.
        :param settings: Configuration.
        :param target_path: Destination path.
        """
        self.root = root
        self.settings = settings
        self.collected = defaultdict(dict)

    def get_permanent_path(self, root: str, static_view_name: str, relative_path: str, hash: str) -> str:
        """Return the permanent path for an asset.

        :param root: Root path.
        :param static_view_name: Name of the asset.
        :param relative_path: Relative path of the asset.
        :param hash: Hash of the file.
        :return: Permanent path to the asset.
        """
        base_file, ext = os.path.splitext(relative_path)
        ext = '.{hash}{ext}'.format(hash=hash, ext=ext)
        relative_path = base_file + ext
        return os.path.join(self.root, MARKER_FOLDER, relative_path)

    def process(self, root: str, static_view_name: str, entry: DirEntry, relative_path: str) -> str:
        """Make a persistent copy of a file.

        :param root: Root path.
        :param static_view_name: Asset name.
        :param entry: DirEntry for asset.
        :param relative_path: Relative path of the asset.
        :return: Permanent path to the asset
        """
        hash = md5(entry.path)
        target = self.get_permanent_path(root, static_view_name, relative_path, hash)
        rel_target = os.path.relpath(target, self.root)

        if os.path.exists(target):
            # Let's avoid unnecessary copy
            if os.path.getsize(target) == entry.stat().st_size:
                # Same size, not a corrupted copy
                return rel_target

        dir = os.path.dirname(target)
        os.makedirs(dir, exist_ok=True)

        # Create a permanent copy
        shutil.copy(entry.path, target)
        logger.info('Writing %s', target)
        return rel_target

    def collect(self, root: str, static_view_name: str, entry: DirEntry, relative_path: str):
        """Process one file and add it to our collection.

        :param root: Root path.
        :param static_view_name: Asset name.
        :param entry: DirEntry for asset.
        :param relative_path: Relative path of the asset.
        """
        target = self.process(root, static_view_name, entry, relative_path)
        by_view = self.collected[static_view_name]
        by_view[relative_path] = target
        self.process(root, static_view_name, entry, relative_path)
        logger.info('Collected %s:%s as %s', static_view_name, relative_path, target)

    def finish(self) -> dict:
        """Finish collection and create manifest.json file.

        :return: Collected files.
        """
        manifest_path = os.path.join(self.root, MARKER_FOLDER, 'manifest.json.tmp')
        dirs = os.path.dirname(manifest_path)
        os.makedirs(dirs, exist_ok=True)
        with open(manifest_path, 'wt') as f:
            json.dump(self.collected, f)
        # Atomic replacement
        # Dotted files should not be accessible through normal static file serving
        os.rename(manifest_path, os.path.join(self.root, MARKER_FOLDER, MANIFEST_FILE))
        return self.collected


class DefaultStaticAssetPolicy(StaticAssetPolicy):
    """Default inplementation of StaticAssetPolicy."""

    def __init__(self, config: Configurator):
        """Initialize DefaultStaticAssetPolicy.

        :param config: Pyramid config.
        """
        super(DefaultStaticAssetPolicy, self).__init__(config)
        self.settings = config.registry.settings

        #: Maintain registry of all registered static views
        self.views = {}

    def add_static_view(self, name: str, path: str):
        """Include a path in static assets and configures cache busting for it.

        This does not only include the static resources in the routing, but sets the default cache busting policy for them in the :term:`production` environment.

        See :py:meth:`pyramid.config.Configurator.add_static_view` and :py:meth:`websauna.system.Initializer.add_cache_buster`

        :param name: Asset name.
        :param path: Asset path.
        """
        # Default value is 0
        cache_max_age = self.settings.get("websauna.cache_max_age_seconds")
        if cache_max_age:
            cache_max_age = int(cache_max_age)

        self.config.add_static_view(name, path, cache_max_age=cache_max_age)

        # If we have caching... we need cachebusters!
        if cache_max_age:
            self.config.add_cache_buster(path, CollectedStaticCacheBuster(name, path, self.settings))

        self.views[name] = path

    def collect_static(self) -> dict:
        """Collect all static files from all static views for the manifest.

        :return: Collect assets.
        """
        results = defaultdict(dict)

        def recurse(collector: CopyAndHashCollector, path: str):
            """Recursively collect assets.

            :param collector: CopyAndHashCollector instance
            :param path: Path to collect assets.
            """
            for entry in scandir(path):
                if entry.name.startswith("."):
                    # Dot files are usually backups or other no no files
                    continue

                # Don't process our internal cache folder
                if MARKER_FOLDER in entry.path:
                    continue

                relative = os.path.relpath(entry.path, collector.root)
                if entry.is_file():
                    collector.collect(path, name, entry, relative)
                elif entry.is_dir():
                    recurse(collector, entry.path)

        r = AssetResolver()
        for name, asset_spec in self.views.items():
            root = r.resolve(asset_spec).abspath()
            collector = CopyAndHashCollector(root, self.settings)
            recurse(collector, root)
            results = collector.finish()

        # Expose for testing
        return results


class CollectedStaticCacheBuster:
    """A Pyramid cache buster which uses persistent static item folder from ws-collect-static command to serve static assets."""

    def __init__(self, static_view_name: str, path: str, settings: dict):
        """Initialize CollectedStaticCacheBuster.

        :param static_view_name: Asset name.
        :param path: Asset path.
        :param settings: Configurations.
        """
        self.settings = settings
        self.static_view_name = static_view_name
        r = AssetResolver()
        self.root = r.resolve(path).abspath()

    @reify
    def manifest(self) -> dict:
        """Read manifest file which maps filenames to their MD5 stamped counterparts.

        :return: Manifest entry for a view.
        """
        target_f = os.path.join(self.root, MARKER_FOLDER, MANIFEST_FILE)
        assert os.path.exists(target_f), "websauna.collected_static manifest does not exist: {path}. Did you run ws-collect-static command?".format(path=target_f)
        with open(target_f, "rt") as f:
            full_manifest = json.load(f)

        # Each view has its own fileset inside cache manifest
        assert self.static_view_name in full_manifest, "Cache manifest did not contain cache data for view {view}, contained {keys}".format(
            view=self.static_view_name,
            keys=full_manifest.keys()
        )
        return full_manifest[self.static_view_name]

    def __call__(self, request: Request, subpath: str, kw: dict) -> t.Tuple[str, dict]:
        """Map a path to perma-asset path.

        :param request: Pyramid request.
        :param subpath: Asset path.
        :param kw: Settings.
        :return: Tuple with perma-asset path and kw.
        """
        perma_subpath = self.manifest.get(subpath, subpath)
        return perma_subpath, kw
